from findCPcore.AbstractMetabolicModel import AbstractMetabolicModel
from findCPcore.State import State
from findCPcore.State import CobraMetabolicStateBuilder

from enum import Enum
import cobra

from math import isnan

from cobra.io.sbml import validate_sbml_model
from cobra.flux_analysis import flux_variability_analysis
from cobra.manipulation import delete
from cobra.manipulation.delete import find_gene_knockout_reactions
from cobra.flux_analysis.deletion import single_reaction_deletion
from cobra.flux_analysis import find_essential_reactions

from optlang.exceptions import SolverError

import itertools
import sys
import threading
import inspect
import ctypes

from time import sleep
from multiprocessing import Process
from sys import platform

# On windows, when deployed with pyinstaller, FVA and other methods get blocked on multiprocessing.
if platform == "win32":
	PROCESSES = 1
else:
	PROCESSES = None

CONST_EPSILON = 0.000005

class NullDevice():
	def write(self, s):
		pass


class CobraMetabolicModel(AbstractMetabolicModel):
	""" Class containing a metabolic network.
		It implements AbstractMetabolicModel with the cobrapy library

	Attributes
		__cobra_model (): Cobra model of the metabolic network. Class: cobra.core.model.

		__objective_value(): float that saves the result of running Flux Balance Analysis on the model.
			The value initially is None.

		__dem (): Dict containing the dead end metabolites of the model.
			- Key: compartments of the model.
			- Values: list of cobra.core.metabolites containing the dead end metabolites of the compartment.
			If the method find_dem() hasnt been called the value of the attribute is None.

		__chokepoints (): List of objects _MetaboliteReact containing chokepoint reactions of the network.
			If the methos find_chokepoints() hasn't been called the value is None.

		__fva (): List of tuples (cobra.core.reaction, maximun, minimum) containing the result of the Flux Variability Analysis.

		__essential_genes (): List of cobra.core.gene containing the essential genes of the model.
		__essential_genes_reactions (): Dict containing the reactions associated to the essential genes.
			Key: cobra.core.reaction.
			Value: list of cobra.core.gene containing the genes associated to the reaction.
		__essential_reactions (): dict: key: reaction id, value: float or "infeasible"

		__spreadsheet (): xlwt Workbook
	"""

	__cobra_model = None
	__objective_value = None
	__dem = None
	__chokepoints = None
	__fva = None
	__essential_genes = None
	__essential_genes_reactions = None
	__essential_reactions = None
	__states = {}


	__exchange_demand_reactions = {}

	# as we consider reactions with flux (0,0) forward (->) save the original topological
	# direction to avoid a direction swap ((<-) to (->)).
	__initial_backward_reactions = set({})


	def model(self):
		return self.__cobra_model

	def set_state(self, key):
		builder = CobraMetabolicStateBuilder()
		self.__states[key] = builder.buildState(self)

	def get_state(self, key):
		if key not in self.__states:
			raise Exception("Couldn't find state: '" + key + "'. State doesn't exist")
		return self.__states[key]

	def model(self):
		return self.__cobra_model

	def id(self):
		return self.__cobra_model.id

	def objective(self):
		return str(self.__cobra_model.objective.expression)

	def objective_value(self):
		return self.__objective_value

	def set_objective(self, reaction):
		""" Sets the objective function of the cobra model as the reactions passed as parameter.

		:param reaction: reaction id of a cobra.core.reaction object
		:type reaction: str
		:raises: RuntimeError: if the reaction id cant be found on the cobra model.
		"""
		try:
			reaction_obj = self.__cobra_model.reactions.get_by_id(reaction)
			self.__cobra_model.objective = str(reaction_obj.id)
		except KeyError as error:
			raise RuntimeError("Reaction id can't be found on the model.")

	def compartments(self):
		""" Return list with compartments

		:return: list with compartments short name
		:rtype: list of str
		"""
		return list(self.__cobra_model.compartments.keys())


	def reactions(self):
		""" Return list with reactions

		:return: list with the reactions of the model
		:rtype:  list of cobra.core.reaction
		"""
		return self.__cobra_model.reactions


	def metabolites(self):
		""" Return list with metabolites

		:return: list with the metabolites of the model
		:rtype:  list of cobra.core.metabolite
		"""
		return self.__cobra_model.metabolites


	def dem(self):
		""" Returns a dict with compartments and its dead-end metabolites if they have been calculated. Dict has
			the following structure:
				- key: compartment name (can be obtained with 'compartments()' method).
				- value: list with objects cobra.core.metabolite that represent the dead-end of the compartment.

		:return: dict with dead-end metabolite by compartment.
		:rtype: dict({ str : list([cobra.core.metabolites]) })
		:raises: Exception: if dead-end hasnt been calculated.
		"""
		if self.__dem is not None:
			return self.__dem
		else:
			raise Exception("Dead End metabolites hasn't been calculated. Please run 'find_dem()'.")


	def chokepoints(self):
		""" Returns a list of tuples of (cobra.core.reaction, cobra.core.metabolite) containing the chokepoint reactions of the model.

		:return: list of tuples (reaction, metabolite) of chokepoint reactions and the metabolite they consume/produce.
		:rtype: list( tuple( cobra.core.reaction, cobra.core.metabolite ) )
		"""
		if self.__chokepoints is not None:
			return [(mtb_rct.reaction, mtb_rct.metabolite) for mtb_rct in self.__chokepoints]
		else:
			raise Exception("Chokepoint reactions hasn't been calculated. Please run 'find_chokepoints()'.")


	def get_fva(self):
		""" Returns the result of running Flux Variability Analysis on the model

		:return: List of tuples (cobra.core.reaction, maximum flux, minimum flux) with the reactions and its flux.
		:rtype: list([ tuple( (cobra.core.reaction, float, float) ) ])
		"""
		if self.__fva is not None:
			return self.__fva
		else:
			raise Exception("Flux Variability hasn't been run. Please run 'fva()'.")


	def genes(self):
		return self.__cobra_model.genes

	def essential_genes(self):
		if self.__essential_genes is not None:
			return self.__essential_genes
		else:
			raise Exception("Essential genes hasn't been calculated. Please run 'find_essential_genes_1()'.")

	def essential_genes_reactions(self):
		if self.__essential_genes_reactions is not None:
			return self.__essential_genes_reactions
		else:
			raise Exception("Essential genes reactions hasn't been calculated. Please run 'find_essential_genes_reactions()'.")

	def essential_reactions(self):
		if self.__essential_reactions is not None:
			return self.__essential_reactions
		else:
			raise Exception("Essential reactions hasn't been calculated. Please run 'find_essential_reactions_1()'.")

	def reversible_reactions(self):
		""" Returns a list of reversible reactions of the model

		:return: List of cobra.core.reaction with reversible direction.
		:rtype: list([ cobra.core.reaction ])
		"""
		return list(filter(lambda x: self.__reaction_direction(x) == self._Direction.REVERSIBLE, self.__cobra_model.reactions))

	def dead_reactions(self):
		""" Returns a list of reactions with upper bound = 0 AND lower bound = 0

		:return: List of cobra.core.reaction with dead-reactions
		:rtype: list([ cobra.core.reaction ])
		"""
		return list(filter(lambda reaction: abs(reaction.upper_bound) < CONST_EPSILON and abs(reaction.lower_bound) < CONST_EPSILON, self.__cobra_model.reactions))

	def exchange_demand_reactions(self):
		return self.__exchange_demand_reactions

	def __id(self, e):
		return e.id


	class _Direction(Enum):
		FORWARD = 0
		BACKWARD = 1
		REVERSIBLE = 2


	def __init__(self, path=None):
		if path is not None:
			self.read_model(path)


	def __print_errors(self, errors):
		for elem in errors:
			aux = errors.get(elem)
			if len(aux) > 0:
				print(elem)
				for error in aux:
					print("    ", error)


	def read_model(self, path):
		""" Reads a cobra model from a file. Assigns it to __cobra_model attribute.

		Args:
			path (): File direction of the cobra model

		:raises RuntimeError: if the input file format is not .xml, .json or .yml.
		:raises FileNotFoundError: if the input file can't be found.
		:raises Exception: if cobrapy throws an exception reading the model (model has errors).
		"""
		original_stderr = sys.stderr  # keep a reference to STDERR
		sys.stderr = NullDevice()  # redirect the real STDERR
		try:
			# Default interface
			cobra.core.model.configuration.solver = 'glpk'

			# check if file exists
			open(path, 'r')
			# read model
			if path[-4:] == ".xml":
				self.__cobra_model = cobra.io.read_sbml_model(path)
			elif path[-5:] == ".json":
				self.__cobra_model = cobra.io.load_json_model(path)
			elif path[-4:] == ".yml":
				self.__cobra_model = cobra.io.load_yaml_model(path)
			else:
				sys.stderr = original_stderr  # turn STDERR back on
				raise RuntimeError("Model file must be either .xml .json .yml")

			# Set default solver: glpk
			self.__cobra_model.solver = 'glpk'

			# Generate exchange/demand reactions list
			exchanges = self.__cobra_model.exchanges
			demands = self.__cobra_model.demands
			self.__exchange_demand_reactions = set(exchanges).union(set(demands))
			for reaction in self.__cobra_model.reactions:
				if (len(reaction.reactants) == 0 and len(reaction.products) == 1) or (len(reaction.reactants) == 1 and len(reaction.products) == 0):
					self.__exchange_demand_reactions.add(reaction)
				if self.__reaction_direction(reaction) == self._Direction.BACKWARD:
					self.__initial_backward_reactions.add(reaction)

		except FileNotFoundError:
			sys.stderr = original_stderr  # turn STDERR back on
			raise FileNotFoundError("File not found: '", path, "'")
			#exit(1)
		except RuntimeError as errorFormat:
			# Formato del fichero incorrecto
			raise(errorFormat)
		except Exception as exception:
			sys.stderr = original_stderr  # turn STDERR back on
			# raise Exception(exception)
			# exit(1)
			(model, errors) = validate_sbml_model(path)
			for code in ['SBML_ERROR', 'SBML_SCHEMA_ERROR', 'SBML_FATAL', 'COBRA_FATAL', 'COBRA_ERROR']:
				error_list = errors[code]
				if error_list != []:
					raise Exception("Error reading model: " + error_list[0])
			# Unexpected error
			raise exception
			

		sys.stderr = original_stderr  # turn STDERR back on


	def save_model(self, path):
		""" Saves the Cobra Model into a file.

		Args:
			path (): Destination file path. Possible file formats:
				- SBML: 	.xml (default)
				- JSON: 	.json
				- YAML: 	.yml
		"""
		if path[-4:] == ".xml":
			cobra.io.write_sbml_model(self.__cobra_model, path)
		elif path[-5:] == ".json":
			cobra.io.save_json_model(self.__cobra_model, path)
		elif path[-4:] == ".yml":
			cobra.io.save_yaml_model(self.__cobra_model, path)
		else:
			raise RuntimeError("Destination file must be either .xml .json .yml")


	def reaction_direction(self, reaction):
		return self.__reaction_direction(reaction)

	def __reaction_direction(self, reaction):
		""" Checks the direction of a reaction given his upper and lower bounds.

		Args:
			reaction (): Class cobra.core.reaction

		Returns: Enum: FORWARD/BACKWARD/REVERSIBLE depending on the direction of the reaction.

		"""
		if reaction.upper_bound > - CONST_EPSILON and reaction.upper_bound < CONST_EPSILON:
			upper = 0
		else:
			upper = reaction.upper_bound
		if reaction.lower_bound > - CONST_EPSILON and reaction.lower_bound < CONST_EPSILON:
			lower = 0
		else:
			lower = reaction.lower_bound
		if lower == 0 and upper == 0:
			if reaction in self.__initial_backward_reactions:
				return self._Direction.BACKWARD
			else:
				return self._Direction.FORWARD
		elif lower >= 0:
			return self._Direction.FORWARD
		elif upper > 0:
			return self._Direction.REVERSIBLE
		else:
			return self._Direction.BACKWARD


	def __is_dead_reaction(self, reaction):
		return abs(reaction.upper_bound) < CONST_EPSILON and abs(reaction.lower_bound) < CONST_EPSILON


	def find_dem_2(self, compartment="ALL"):
		""" Finds the dead end metabolites of the model or a specific comparment.
			When searched, it saves them in the '__dem' class atribute with the compartment as a key and
			a list of cobra.core.metabolite as value representing the DEM of that compartment.

		:param compartment: String representing the compartment in which the search is made. "ALL" by default
		:type compartment: str
		:return: dict with dead - end metabolites by compartment.
		:rtype: dict({str: list([cobra.core.metabolites])})
		"""
		if self.__dem == None:
			self.__dem = {}
		list_dem = []
		for metabolite in self.__cobra_model.metabolites:
			# If the metabolite is in the choosen compartment. Else we ignore it
			if (metabolite.compartment == compartment) or (compartment == "ALL"):
				# If metabolite only appears in one reaction and it's not reversible -> dead end
				if len(metabolite.reactions) == 1:
					reaction = list(metabolite.reactions)[0]
					if self.__reaction_direction(reaction) != self._Direction.REVERSIBLE:
						list_dem = list_dem + [metabolite]
				# If metabolite appears in more than one reacion
				# Check if it's always product or reactant
				else:
					reactions = list(metabolite.reactions)
					i = 0
					end = False
					reactant = None 	# variable for comparing each reaction with the previous one
					while i < len(reactions) and not end:
						reaction = reactions[i]
						# If reaction is reversible -> not dead end
						if self.__reaction_direction(reaction) == self._Direction.REVERSIBLE:
							end = True
						else:
							# first time entering
							if reactant == None:

								if (self.__reaction_direction(reaction) == self._Direction.FORWARD and metabolite in reaction.reactants) \
									or (self.__reaction_direction(reaction) == self._Direction.BACKWARD and metabolite in reaction.products):
									reactant = True
								else:
									reactant = False
							# not first time -> 'reactant' already assigned
							else:
								if (self.__reaction_direction(reaction) == self._Direction.FORWARD and metabolite in reaction.reactants) \
									or (self.__reaction_direction(reaction) == self._Direction.BACKWARD and metabolite in reaction.products):
									if not reactant:
										end = True
								else:
									if reactant:
										end = True
						i = i + 1
					if not end:
						list_dem = list_dem + [metabolite]
		if compartment != "ALL":
			self.__dem[compartment] = list_dem
		else:
			for comprt in self.__cobra_model.compartments:
				self.__dem[comprt] = []
			for metabolite in list_dem:
				self.__dem[metabolite.compartment].append(metabolite)
		return list_dem


	def find_dem(self, compartment="ALL"):
		""" Finds the dead end metabolites of the model or a specific comparment.
			When searched, it saves them in the '__dem' class atribute with the compartment as a key and
			a list of cobra.core.metabolite as value representing the DEM of that compartment.

		:param compartment: String representing the compartment in which the search is made. "ALL" by default
		:type compartment: str
		:return: dict with dead - end metabolites by compartment.
		:rtype: dict({str: list([cobra.core.metabolites])})
		"""
		if self.__dem == None:
			self.__dem = {}
		reactants = set({})
		products = set({})
		for reaction in self.__cobra_model.reactions:
			direction = self.__reaction_direction(reaction)
			if direction == self._Direction.REVERSIBLE:
				# All metabolites are reactants and products
				for metabolite in reaction.metabolites:
					if (compartment == "ALL") or (metabolite.compartment == compartment):
						reactants.add(metabolite)
						products.add(metabolite)
			elif direction == self._Direction.BACKWARD:
				# If reaction is backward -> products are reactants and reactants are products
				for metabolite in reaction.reactants:
					if (compartment == "ALL") or (metabolite.compartment == compartment):
						products.add(metabolite)
				for metabolite in reaction.products:
					if (compartment == "ALL") or (metabolite.compartment == compartment):
						reactants.add(metabolite)
			else:
				# Forward reaction
				for metabolite in reaction.reactants:
					if (compartment == "ALL") or (metabolite.compartment == compartment):
						reactants.add(metabolite)
				for metabolite in reaction.products:
					if (compartment == "ALL") or (metabolite.compartment == compartment):
						products.add(metabolite)
		# Diferrence beetwen sets
		dem_reactants = reactants.difference(products)
		dem_products = products.difference(reactants)
		# All dem = sum of dem produced and dem consumed
		dem = list(dem_reactants) + list(dem_products)
		if compartment != "ALL":
			self.__dem[compartment] = dem
		else:
			for comprt in self.__cobra_model.compartments:
				self.__dem[comprt] = []
			for metabolite in dem:
				self.__dem[metabolite.compartment].append(metabolite)
		return dem


	def __check_dem(self, compartment="ALL"):
		""" Checks if the dead end metabolites of the model or a specific compartmet has been searched.
			If not searches the dead end metabolites.

		Args:
			compartment (): limit the checkup to a specific compartment. "ALL" by default
		"""
		if (compartment == "ALL"):
			if (self.__dem == None):
				self.find_dem()
			else:
				for cmprt in self.__cobra_model.compartments:
					if cmprt not in self.__dem:
						self.find_dem(cmprt)
		elif (self.__dem == None) or (compartment not in self.__dem):
			self.find_dem(compartment)


	class _MetaboliteReact:
		""" Class composed by a metabolite and a reaction containing it.

		Attributes
			metabolite : Class cobra.core.metabolite
			reaction : Class cobra.core.reaction
		"""

		def __init__(self, metabolite, reaction):
			""" Private class initiator.
			"""
			self.metabolite = metabolite
			self.reaction = reaction


	def __metabolite_id(self, obj):
		""" Given an object _MetaboliteReact it returns the 'id' of the metabolite

		Args:
			obj (): Object cobra.core.metabolite

		Returns: 'id' field of the metabolite

		"""
		return obj.metabolite.id


	def __reaction_id(self, obj):
		""" Given an object _MetaboliteReact it returns the 'id' of the reaction

		Args:
			obj (): Object cobra.core.reaction

		Returns: 'id' field of the reaction

		"""
		return obj.reaction.id


	def find_chokepoints(self, exclude_dead_reactions=False):
		""" Finds chokepoint reactions of the cobra model

		Returns: List of objects of class _MetaboliteReact containing a chokepoint reaction
				 and the metabolite consumed/produced

		"""
		reactants = []	# List of reactants metabolites with it's reaction
		products = []	# List of products metabolites with it's reaction
		for reaction in self.__cobra_model.reactions:
			direction = self.__reaction_direction(reaction)
			is_dead_reaction = self.__is_dead_reaction(reaction)
			if (not exclude_dead_reactions) or (exclude_dead_reactions and not is_dead_reaction):
				if direction == self._Direction.REVERSIBLE:
					# If reaction is reversible all metabolites are reactants and products
					for metabolite in reaction.metabolites:
						mtb = self._MetaboliteReact(metabolite, reaction)
						reactants.append(mtb)
						products.append(mtb)
				else:
					if direction == self._Direction.FORWARD:
						reac_reactants = reaction.reactants
						reac_prodcuts = reaction.products
					else:
						# If reaction is backward reactants=products and products=reactants
						reac_reactants = reaction.products
						reac_prodcuts = reaction.reactants
					# add reactants and products of the reaction to the list
					for metabolite in reac_prodcuts:
						mtb = self._MetaboliteReact(metabolite, reaction)
						products.append(mtb)
					for metabolite in reac_reactants:
						mtb = self._MetaboliteReact(metabolite, reaction)
						reactants.append(mtb)

		# Order lists
		reactants.sort(key=self.__metabolite_id)
		products.sort(key=self.__metabolite_id)

		i = 0
		j = 0
		chokepoints = []

		# Pair metabolites of ordered lists of reactants and products
		# 	If a metabolite appears 1 time in one list and more than 1 time on the other -> 
		#	The reaction on the 1 time side is a chokepoint reaction
		while i < len(reactants) and j < len(products):
			mtb1 = reactants[i].metabolite.id
			mtb2 = products[j].metabolite.id
			if mtb1 < mtb2:
				i = i + 1
			elif mtb1 > mtb2:
				j = j + 1
			else:  # mtb1 = mtb2	Metabolites are the same
				# Pairing the 2 metabolites
				num1 = 1 # Number of times de first metabolite appears
				num2 = 1 # Number of times the second metabolite appears
				i = i + 1
				j = j + 1
				while i < len(reactants) and reactants[i].metabolite.id == mtb1:
					num1 = num1 + 1
					i = i + 1
				while j < len(products) and products[j].metabolite.id == mtb2:
					num2 = num2 + 1
					j = j + 1
				#if num1 == 1 and num2 >= 1:
				if num1 == 1:
					chokepoints.append(reactants[i - 1])
					if num2 == 1:
						chokepoints.append(products[j - 1])
				#elif num2 == 1 and num1 >= 1:
				elif num2 == 1:
					chokepoints.append(products[j - 1])
					if num1 == 1:
						chokepoints.append(reactants[i - 1])
		self.__chokepoints = chokepoints
		return chokepoints


	def __remove_dem(self, delete_exchange=False, keep_all_incomplete_reactions=True):
		""" Auxiliar method for 'remove_dem()'

		"""
		while True:
			dem = list(itertools.chain.from_iterable(self.__dem.values()))
			num_mtbs = len(self.__cobra_model.metabolites)
			self.__cobra_model.remove_metabolites(dem)
			reactions = []
			# Delete the reactions that doesnt produce or doesnt consume
			if delete_exchange:
				for reaction in self.__cobra_model.reactions:
					if len(reaction.reactants) == 0 or len(reaction.products) == 0:
						reactions.append(reaction)
			else:
				if keep_all_incomplete_reactions:
					for reaction in self.__cobra_model.reactions:
						if (len(reaction.reactants) == 0 or len(reaction.products) == 0) and (reaction not in self.__exchange_demand_reactions):
							reactions.append(reaction)
				else:
					exchange_and_demand = set(exchanges).union(set(demands))
					for reaction in self.__cobra_model.reactions:
						if (len(reaction.reactants) == 0 or len(reaction.products) == 0) and (reaction not in exchange_and_demand):
							reactions.append(reaction)
			self.__cobra_model.remove_reactions(reactions)
			self.find_dem()
			# Loop continues while the number of metabolites doenst change
			if num_mtbs == len(self.__cobra_model.metabolites):
				break


	def remove_dem(self, delete_exchange=False, keep_all_incomplete_reactions=True):
		""" While there network changes, eliminates dead ends metabolites
			and reactions that only produce or consume

			- delete_exchange:
				- True: all the reactions that are produce or consume 0 metabolites are deleted whether they are exchange/demand or not.
				- False: deleted according to 'keep_all_incomplete_reactions' param.
			- keep_all_incomplete_reactions:
				- False: if a reactions is in [cobra Boundary reactions](https://cobrapy.readthedocs.io/en/latest/media.html#Boundary-reactions) (calculated by heuristics) that reaction can't be deleted.
				- True: if a reaction initially doesn't produce or consume any metabolite that reaction can't be deleted.

		:param delete_exchange: if True exchange and demand reactions are deleted
		:type delete_exchange: bool
		:param keep_all_incomplete_reactions: If True all reactions that initially dont consume or dont produce any
				metabolite are kept.
		:type keep_all_incomplete_reactions: bool
		:return:
		:rtype:
		"""

		self.__check_dem()

		if delete_exchange == True:
			self.__remove_dem(True, keep_all_incomplete_reactions)
		elif len(self.__cobra_model.exchanges) == 0:
			self.__remove_dem(True, keep_all_incomplete_reactions)
		else:
			self.__remove_dem(False, keep_all_incomplete_reactions)

		""" 

		Args:
			loopless (): Runs a lopeless analysis
			verbose (): Prints the results obtained in the analysis
			update_flux (): Updates the bounds of the reaction with the values obtained with the F.V.A.
		"""
	def fva(self, loopless=False, verbose=False, update_flux=False, threshold=None, pfba_factor=None):
		""" If possible, runs a Flux Variability Analysis on the model and saves the result on the '__fva' class atribute.
			Returns a list of errors. If there wasn't any error while running FVA it return an empty list: []

			For more info about the params see: https://cobrapy.readthedocs.io/en/latest/autoapi/cobra/flux_analysis/variability/index.html?highlight=flux_varia#cobra.flux_analysis.variability.flux_variability_analysis

		:param loopless: Runs a lopeless analysis
		:type loopless: bool
		:param verbose: Print the result of FVA while running the analysis.
		:type verbose: bool
		:param update_flux: Updates the bounds of the reaction with the values obtained with the F.V.A.
		:type update_flux: bool
		:param threshold: Must be <= 1.0. If is None: deafult = 1.0. factor of the maximum objective value.
		:type threshold: float
		:param pfba_factor: Add an additional constraint to the model that requires the total sum of absolute fluxes
				must not be larger than this value times the smallest possible sum of absolute fluxes
		:type pfba_factor: float
		:return: list of errors if there was any. Else return an empty list: []
		:rtype: list([ str ])
		"""
		errors = []
		if verbose:
			print("FLUX VARIABILITY ANALYSIS: ",  self.__cobra_model.id)
		try:
			i = 0
			fva_result = []
			# Run FVA analysis
			if threshold is None and pfba_factor is None:
				fva_reactions = flux_variability_analysis(self.__cobra_model, loopless=loopless, processes=PROCESSES)
			else:
				fva_reactions = flux_variability_analysis(self.__cobra_model, loopless=loopless, fraction_of_optimum=threshold,  pfba_factor=pfba_factor, processes=PROCESSES)
			for reaction_id in fva_reactions.index:
				reaction = self.__cobra_model.reactions.get_by_id(reaction_id)
				fva_lower = float(fva_reactions.values[i][0])
				fva_upper = float(fva_reactions.values[i][1])
				# Print info
				if verbose:
					print("REACTION: ", reaction.name)
					if update_flux:
						print("    model bounds: [", str(reaction.upper_bound)[:10].ljust(10), ", ", str(reaction.lower_bound)[:10].ljust(10), "]")
					print("    fva ranges:   [", str(fva_upper)[:10].ljust(10), ", ", str(fva_lower)[:10].ljust(10), "]")
				# Update bounds
				if update_flux:
					reaction.lower_bound = fva_lower
					reaction.upper_bound = fva_upper
				i = i + 1
				# Add results to list
				fva_result.append( (reaction, fva_upper, fva_lower)  )
			self.__fva = fva_result
		except cobra.exceptions.Infeasible as error:
			self.__fva = []
			errors.append("Model is infeasable")
		except Exception as error:
			self.__fva = []
			if "unbounded" in str(error):
				errors.append("Model is unbounded")
			else:
				errors.append(str(error))
		return errors


	def get_growth(self):
		""" Runs flux balance analysis (slim_optimize() from cobrapy) and returns the objective value.
			Saves the objective value in __objective_value attribute.

		:return: objective value
		:rtype: float
		"""
		model = self.__cobra_model
		try:
			if 'moma_old_objective' in model.solver.variables:
				model.slim_optimize()
				growth = model.solver.variables.moma_old_objective.primal
			else:
				growth = model.slim_optimize()
			self.__objective_value = growth
		except SolverError:
			growth = float('nan')
		except Exception as timeout:
			growth = float('-1')
		return growth


	def find_essential_reactions_1(self):
		errors = []
		try:
			self.__objective_value = self.get_growth()
			if isnan(self.__objective_value):
				self.__objective_value = None
			self.__essential_reactions = {}
			deletions = single_reaction_deletion(self.__cobra_model, method='fba', processes=PROCESSES)
			essential = deletions.loc[:, :]['growth']
			for frozset in essential.index.values:
				reaction = self.__cobra_model.reactions.get_by_id(list(frozset)[0])
				self.__essential_reactions[reaction] = essential[frozset]
		except Exception as error:
			errors.append(str(error))
			self.__essential_reactions = {}
		return errors


	def find_essential_genes_1(self):
		""" If the model has genes, finds the essential genes and assigns them to the '__essential_genes' class atribute.

		"""
		errors = []
		try:
			# TODO: check processes number (because deadlock)
			aux = cobra.flux_analysis.variability.find_essential_genes(self.__cobra_model, processes=PROCESSES)
			self.__essential_genes = aux
		except cobra.exceptions.Infeasible as error:
			self.__essential_genes = []
			errors.append("Model is infeasable")
		except Exception as error:
			self.__essential_genes = []
			if "unbounded" in str(error):
				errors.append("Model is unbounded")
			else:
				errors.append(str(error))
		return errors

	def find_essential_genes_reactions(self):
		""" If the model has genes finds the reactions associated with the essential genes.
			Searches essential genes if hasn't done before.

		"""
		errors = []
		try:
			if self.__essential_genes is not None:
				# Dict with reaction as key and list of genes as value
				reactions = {}
				for gene in self.__essential_genes:
					reactions_knock = find_gene_knockout_reactions(self.__cobra_model, [gene])
					for reaction in reactions_knock:
						if reaction in reactions:
							reactions[reaction].append(gene)
						else:
							reactions[reaction] = [gene]
				self.__essential_genes_reactions = reactions
		except Exception as error:
			self.__essential_genes_reactions = {}
			errors.append(str(error))
		return errors


	def print_model_info(self):
		""" Prints model general info

		"""
		print("MODEL INFO")
		print('-' * 55)
		print("MODEL: ", self.__cobra_model.id)
		print("REACTIONS: ", len(self.__cobra_model.reactions))
		print("METABOLITES: ", len(self.__cobra_model.metabolites))
		print("GENES: ", len(self.__cobra_model.genes))
		if len(self.__cobra_model.compartments):
			print("COMPARTMENTS: ",  list(self.__cobra_model.compartments)[0])
			i = 1
			while i < len(self.__cobra_model.compartments):
				print("".ljust(14), list(self.__cobra_model.compartments)[i])
				i = i + 1
		print()

	def print_metabolites(self, ordered=False):
		""" Prints metabolites and its reactions of the cobra model

			Args:
				ordered (): print the metabolites in alphabetical order by id

		"""
		metabolites = self.__cobra_model.metabolites
		if ordered:
			metabolites.sort(key=self.__id)
		print("MODEL: ", self.__cobra_model.id, " - NUMBER OF METABOLITES: ", len(self.__cobra_model.metabolites))
		print("METABOLITE".ljust(10), " | ", "COMPARTMENT".ljust(15), " | ", "REACTION ID")
		print('-' * 55)
		for metabolite in metabolites:
			print(metabolite.id.ljust(10), " | ", metabolite.compartment.ljust(15), " | ",list(metabolite.reactions)[0].id)
			i = 1
			while i < len(list(metabolite.reactions)):
				print("".ljust(10), " | ", "".ljust(15), " | ", list(metabolite.reactions)[i].id)
				i = i + 1
		print()


	def print_reactions(self, ordered=False):
		""" Prints reactions of the cobra model

			Args:
				ordered (): print the reactions in alphabetical order by id

		"""
		reactions = self.__cobra_model.reactions
		if ordered:
			reactions.sort(key=self.__id)
		print("MODEL: ", self.__cobra_model.id, " - NUMBER OF REACTIONS: ", len(self.__cobra_model.reactions))
		print("REACTION ID | UPPER BOUND | LOWER BOUND | REACTION")
		print('-' * 55)
		for reaction in reactions:
			print(reaction.id.ljust(10), " |  ", str(reaction.upper_bound)[:8].ljust(8), " | ", str(reaction.lower_bound)[:9].ljust(9), " | ", reaction.reaction)
		print()


	def print_genes(self, ordered=False):
		""" If the model has genes prints the essential genes of the model. If essential genes hasn't been searched it
			searches them.

		Args:
			ordered (): print the essential genes in alphabetical order by id

		"""
		genes = list(self.__cobra_model.genes)
		print("MODEL: ", self.__cobra_model.id, " - NUMBER OF GENES: ", len(genes))
		print("GENE ID".ljust(10), " | ", "GENE NAME".ljust(10), " | ", "REACTION ID".ljust(10), "| ", "GPR RELATION")
		print('-' * 55)
		if ordered:
			genes.sort(key=self.__id)
		for gene in genes:
			reactions = list(gene.reactions)
			print(gene.id.ljust(10), " | ", gene.name.ljust(10), " | ",  reactions[0].id.ljust(10), " | ",  reactions[0].gene_reaction_rule.ljust(15))
			i = 1
			while i < len(reactions):
				print("".ljust(10), " | ", "".ljust(10), " | ", reactions[i].id.ljust(10), " | ",  reactions[i].gene_reaction_rule.ljust(15))
				i = i + 1
		print()


	def print_dem(self, ordered=False, compartment="ALL"):
		""" Prints dead end metabolites of the cobra model

			Args:
				ordered (): print the dead end metabolites in alphabetical order by id
				compartment (): show the dead end metabolites of a specific compartment. "ALL" by default.
		"""
		self.__check_dem(compartment)
		if compartment == "ALL":
			metabolites = list(itertools.chain.from_iterable(self.__dem.values()))
		else:
			metabolites = self.__dem[compartment]
		if ordered:
			metabolites.sort(key=self.__id)
		print("MODEL: ", self.__cobra_model.id, " - NUMBER OF DEM: ", len(self.__dem), " - COMPARTMENT: ", compartment)
		print("METABOLITE".ljust(10), " | ", "COMPARTMENT".ljust(15), " | ", "REACTION ID")
		print('-' * 55)
		for metabolite in metabolites:
			print(metabolite.id.ljust(10), " | ", metabolite.compartment.ljust(15), " | ",list(metabolite.reactions)[0].id)
			i = 1
			while i < len(list(metabolite.reactions)):
				print("".ljust(10), " | ", "".ljust(15), " | ", list(metabolite.reactions)[i].id)
				i = i + 1
		print()


	def print_chokepoints(self, ordered=False):
		""" Prints chokepoints reactions of the cobra model and its consumed/produced metabolites.

			Args:
				ordered (): print the chokepoint reactions in alphabetical order by id

		"""
		if self.__chokepoints is None:
			self.find_chokepoints()
		chokepoints = self.__chokepoints
		if ordered:
			chokepoints.sort(key=self.__reaction_id)
		print("MODEL: ", self.__cobra_model.id, " - NUMBER OF CHOKEPOINTS: ", len(self.__chokepoints))
		print("METABOLITE ID | ", "METABOLITE NAME".ljust(40), " | REACTION ID | REACTION NAME")
		print('-' * 60)
		for mtb_rct in chokepoints:
			print(mtb_rct.metabolite.id.ljust(12), " | ", mtb_rct.metabolite.name.ljust(40), " | ", mtb_rct.reaction.id.ljust(9)," | ", mtb_rct.reaction.name)
		print()



	def print_essential_genes(self, ordered=False):
		""" If the model has genes prints the essential genes of the model. If essential genes hasn't been searched it
			searches them.

		Args:
			ordered (): print the essential genes in alphabetical order by id

		"""
		errors = []
		if self.__essential_genes is None:
			#print("Essential genes hasn't been searched")
			#print("Searching essential genes...")
			errors = self.find_essential_genes_1()
		if errors != []:
			print(errors[0])
		else:
			genes = list(self.__essential_genes)
			print("MODEL: ", self.__cobra_model.id, " - NUMBER OF GENES: ", len(genes))
			print("GENE ID".ljust(10), " | ", "GENE NAME".ljust(10), " | ", "REACTION ID".ljust(10), "| ","GPR RELATION")
			print('-' * 55)
			if ordered:
				genes.sort(key=self.__id)
			for gene in genes:
				reactions = list(gene.reactions)
				print(gene.id.ljust(10), " | ", gene.name.ljust(10), " | ", reactions[0].id.ljust(10), " | ",reactions[0].gene_reaction_rule.ljust(15))
				i = 1
				while i < len(reactions):
					print("".ljust(10), " | ", "".ljust(10), " | ", reactions[i].id.ljust(10), " | ",reactions[i].gene_reaction_rule.ljust(15))
					i = i + 1
			print()


	def print_essential_genes_reactions(self, ordered=False):
		""" If the model has genes prints the reactions associated to the essential genes of the model.
			If essential genes hasn't been searched it searches them.

			Args:
				ordered (): print the reactions in alphabetical order by id

			"""
		errors = []
		if self.__essential_genes_reactions is None:
			#print("Essential genes hasn't been searched")
			#print("Searching essential genes...")
			errors = self.find_essential_genes_reactions()
		if errors != []:
			print(errors[0])
		else:
			print("MODEL: ", self.__cobra_model.id, " - NUMBER OF ESSENTIAL GENES REACTIONS: ", len(self.__essential_genes_reactions))
			print("REACTION ID | GENE ID     | GPR RELATION ")
			print('-' * 60)
			reactions = self.__essential_genes_reactions
			reactions_key = reactions.keys()
			if ordered:
				reactions_key = sorted(reactions.keys(), key=self.__id)
			for reaction in reactions_key:
				genes = reactions[reaction]
				print(reaction.id.ljust(10), " | ", genes[0].id.ljust(10), "|", reaction.gene_reaction_rule)
				i = 1
				while i < len(genes):
					print("".ljust(10), " | ", genes[0].id.ljust(10), "|", reaction.gene_reaction_rule)
					i = i + 1
			print()

	def print_essential_reactions(self, ordered=False):
		"""
		"""
		errors = []
		if self.__essential_reactions is None:
			#print("Essential genes hasn't been searched")
			#print("Searching essential genes...")
			errors = self.find_essential_reactions_1()
		if errors != []:
			print(errors[0])
		else:
			print("MODEL: ", self.__cobra_model.id, " - REACTIONS KNOCKOUT RESULT")
			print("REACTION ID     | FBA OBJECTIVE ")
			print('-' * 60)
			reactions = self.__essential_reactions
			reactions_key = reactions.keys()
			if ordered:
				reactions_key = sorted(reactions.keys(), key=self.__id)
			for reaction in reactions_key:
				print(reaction.id.ljust(15), "|", reactions[reaction])
			print()






import xlwt

from findCPcore.MetabolicModel import MetabolicModel
from findCPcore.CobraMetabolicModel import CobraMetabolicModel
from findCPcore.Spreadsheet import Spreadsheet

class ErrorGeneratingModel(Exception):
	pass


class FacadeUtils:

	def run_sensibility_analysis(self, model_path, print_f, arg1, arg2, objective=None):
		s = xlwt.Workbook()
		style = xlwt.XFStyle()
		font = xlwt.Font()
		font.bold = True
		style.font = font
		sheet = s.add_sheet("main")

		FRAC = [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]

		print_f("Reading model...", arg1, arg2)
		MODEL = model_path
		model = CobraMetabolicModel(MODEL)
		if objective is not None:
			model.set_objective(objective)

		reversible_initial = set([r.id for r in list(filter(lambda x: model.reaction_direction(x)==model._Direction.REVERSIBLE, model.reactions()))])
		dead_reactions_initial = set([r.id for r in model.dead_reactions()])
		nr_initial = set([r.id for r in model.reactions()]).difference(dead_reactions_initial).difference(reversible_initial)

		model.find_chokepoints(exclude_dead_reactions=True)
		chokepoints_initial = set({})
		for (reaction, metabolite) in model.chokepoints():
			chokepoints_initial.add(reaction.id)
		chokepoints_initial = chokepoints_initial.difference(dead_reactions_initial)

		sheet.write(0,0, "Model", style=style)
		sheet.write(2,0, "Reactions", style=style)
		sheet.write(4,0, "Metabolites", style=style)	
		sheet.write(1,   0, model.id())
		sheet.write(3,   0, len(model.reactions()))
		sheet.write(5,   0, len(model.metabolites()))

		y = 1
		for i in range(0, len(FRAC)):
		
			print_f("Running Flux Variability Analysis with fraction: " + str(FRAC[i]), arg1, arg2)
			model = CobraMetabolicModel(MODEL)
			if objective is not None:
				model.set_objective(objective)
			errors_fva = model.fva(update_flux=True, threshold=FRAC[i])

			if errors_fva != []:
				print_f("Couldn't run Flux Variability Analysis: " + str(errors_fva[0]), arg1, arg2)
				sheet.write(0, i+4, "gamma = " + str(FRAC[i]), style=style)
				sheet.write(y, i+4, "Error running FVA: " + str(errors_fva[0]))
				sheet.write(y+1, i+4, "Error running FVA: " + str(errors_fva[0]))
				sheet.write(y+2, i+4, "Error running FVA: " + str(errors_fva[0]))
				sheet.write(y+3, i+4, "Error running FVA: " + str(errors_fva[0]))
				continue

			fva_dead_reactions = set([r.id for r in model.dead_reactions()])
			fva_reversible = set([r.id for r in list(filter(lambda x: model.reaction_direction(x)==model._Direction.REVERSIBLE, model.reactions()))])

			nr = set([r.id for r in model.reactions()]).difference(fva_reversible).difference(fva_dead_reactions)

			#nnrr = reversible_initial.difference(fva_reversible).difference(fva_dead_reactions)

			model.find_chokepoints(exclude_dead_reactions=True)
			chokepoints = set({})
			for (reaction, metabolite) in model.chokepoints():
				chokepoints.add(reaction.id)
			test1 = len(chokepoints)
			chokepoints = chokepoints.difference(fva_dead_reactions)
			test2 = len(chokepoints)
			assert(test1 == test2)

			sheet.write(0, i+4, "gamma = " + str(FRAC[i]), style=style)
			sheet.write(y, i+4, len(fva_reversible))
			sheet.write(y+1, i+4, len(nr))
			sheet.write(y+2, i+4, len(fva_dead_reactions))
			sheet.write(y+3, i+4, len(chokepoints))

		sheet.write(y, 2, "Reversible Reactions (RR)", style=style)
		sheet.write(y+1, 2, "Non-reversible Reactions (NR)", style=style)
		sheet.write(y+2, 2, "Dead Reactions (DR)", style=style)
		sheet.write(y+3,   2, "Chokepoints (CP)", style=style)

		sheet.write(0,3, "INITIAL", style=style)
		sheet.write(y, 3, len(reversible_initial))
		sheet.write(y+1, 3, len(nr_initial))
		sheet.write(y+2, 3, len(dead_reactions_initial))
		sheet.write(y+3,   3, len(chokepoints_initial))
		
		y = y + 5

		sObject = Spreadsheet()
		sObject.set_workbook(s)
		return sObject

	def run_summary_model(self, model_path, print_f, arg1, arg2, objective=None, fraction=1.0):
		#verboseprint = print if verbose else lambda *a, **k: None

		print_f("Reading model...", arg1, arg2)
		model = MetabolicModel(CobraMetabolicModel(model_path))
		if objective is not None:
			model.set_objective(objective)

		model.set_state("initial")
		model.set_state("dem")
		model.set_state("fva")
		model.set_state("fva_dem")

		s = Spreadsheet()
		s.spreadsheet_write_model_info(model, "model_info")
		s.spreadsheet_write_reactions(model, "reactions", ordered=True)
		s.spreadsheet_write_metabolites(model, "metabolites", ordered=True, print_reactions=True)
		s.spreadsheet_write_genes(model, "genes", ordered=True, print_reactions=True)

		print_f("Generating models...", arg1, arg2)

		model.find_essential_genes_reactions()
		print_f("Searching Dead End Metabolites (D.E.M.)...", arg1, arg2)
		model.find_dem()
		print_f("Searching chokepoint reactions...", arg1, arg2)
		model.find_chokepoints(exclude_dead_reactions=True)
		print_f("Searching essential reactions...", arg1, arg2)
		model.find_essential_reactions_1()
		print_f("Searching essential genes...", arg1, arg2)
		errors_initial = model.find_essential_genes_1()
		if errors_initial != []:
			MSG = "Couldn't find essential genes: " + str(errors_initial[0])
			print_f(MSG)
		else:
			print_f("Searching essential genes reactions...", arg1, arg2)
			model.find_essential_genes_reactions()

		model.set_state("initial")

		print_f("Removing Dead End Metabolites (D.E.M.)...", arg1, arg2)
		model.remove_dem()
		print_f("Searching essential reactions...", arg1, arg2)
		model.find_essential_reactions_1()
		print_f("Searching new chokepoint reactions...", arg1, arg2)
		model.find_chokepoints(exclude_dead_reactions=True)

		if errors_initial == []:
			print_f("Searching essential genes...", arg1, arg2)
			errors_dem = model.find_essential_genes_1()
			if errors_dem == []:
				print_f("Searching essential genes reactions...", arg1, arg2)
				model.find_essential_genes_reactions()

		model.set_state("dem")

		print_f("Running Flux Variability Analysis...", arg1, arg2)
		model = MetabolicModel(CobraMetabolicModel(model_path))

		#reaction_ob = model.model().reactions.get_by_id("EX_cl_LPAREN_e_RPAREN_")
		#reaction_ob.upper_bound = float(0.0)
		#reaction_ob.lower_bound = float(0.0)

		if objective is not None:
			model.set_objective(objective)
		errors_fva = model.fva(update_flux=True, threshold=fraction)

		if errors_fva != []:
			MSG = "Couldn't run Flux Variability Analysis: " + str(errors_fva[0])
			print_f(MSG, arg1, arg2)
		else:
			print_f("Searching Dead End Metabolites (D.E.M.)...", arg1, arg2)
			model.find_dem()
			print_f("Searching new chokepoint reactions...", arg1, arg2)
			model.find_chokepoints(exclude_dead_reactions=True)
			print_f("Searching essential genes...", arg1, arg2)
			errors_fva_genes = model.find_essential_genes_1()
			if errors_fva_genes != []:
				MSG = "Couldn't find essential genes: " + str(errors_fva_genes[0])
				print_f(MSG)
			else:
				print_f("Searching essential genes reactions...", arg1, arg2)
				model.find_essential_genes_reactions()
			print_f("Searching essential reactions...", arg1, arg2)
			model.find_essential_reactions_1()

			s.spreadsheet_write_reactions(model, "reactions_FVA", ordered=True)
			s.spreadsheet_write_metabolites(model, "metabolites_FVA", ordered=True, print_reactions=True)

			model.set_state("fva")

			print_f("Removing Dead End Metabolites (D.E.M.)...", arg1, arg2)
			model.remove_dem()
			print_f("Searching essential reactions...", arg1, arg2)
			model.find_essential_reactions_1()
			print_f("Searching new chokepoint reactions...", arg1, arg2)
			model.find_chokepoints(exclude_dead_reactions=True)
			if errors_fva_genes == []:
				print_f("Searching essential genes...", arg1, arg2)
				model.find_essential_genes_1()
				print_f("Searching essential genes reactions...", arg1, arg2)
				model.find_essential_genes_reactions()

			model.set_state("fva_dem")

		print_f("Generating spreadsheet...", arg1, arg2)
		s.spreadsheet_write_reversible_reactions("reversible reactions", model.get_state("initial"), model.get_state("fva"), ordered=True)
		s.spreadsheet_write_summary_reactions("chokepoints", model.get_state("initial"), model.get_state("dem"),
		                                      model.get_state("fva"),
		                                      model.get_state("fva_dem"))
		s.spreadsheet_write_summary_metabolites("dead-end", model.get_state("initial"), model.get_state("fva"))
		s.spreadsheet_write_chokepoints_genes("comparison", model.get_state("initial"),
		                                      model.get_state("dem"),
		                                      model.get_state("fva"), model.get_state("fva_dem"))
		s.spreadsheet_write_essential_reactions("essential reactions", model.get_state("initial"),
		                                      model.get_state("dem"),
		                                      model.get_state("fva"), model.get_state("fva_dem"), ordered=True)
		s.spreadsheet_write_summary("summary", model.get_state("initial"),
		                                      model.get_state("dem"),
		                                      model.get_state("fva"), model.get_state("fva_dem"))

		return s

	def find_and_remove_dem(self, model_path):
		model = MetabolicModel(CobraMetabolicModel(model_path))
		model.find_dem()
		model.remove_dem()
		return model

	def run_fva(self, model_path, objective=None, fraction=1.0):
		model = MetabolicModel(CobraMetabolicModel(model_path))
		if objective is not None:
			model.set_objective(objective)
		errors_fva = model.fva(update_flux=True)
		errors = model.fva(update_flux=True, threshold=fraction)
		if errors != []:
			return (model, errors[0])
		else:
			return (model, "")

	def run_fva_remove_dem(self, model_path, objective=None, fraction=1.0):
		model = MetabolicModel(CobraMetabolicModel(model_path))
		if objective is not None:
			model.set_objective(objective)
		errors = model.fva(update_flux=True, threshold=fraction)
		if errors != []:
			return (model, errors[0])
		else:
			model.find_dem()
			model.remove_dem()
			return (model, "")

	def save_model(self, output_path, model):
		if output_path != "":
			try:
				model.save_model(output_path)
				return (True, output_path)
			except Exception as error:

				return (False, str(error))
		return (False, '')

	def save_spreadsheet(self, output_path, spreadsheet):
		if output_path != "":
			try:
				spreadsheet.spreadsheet_save_file(output_path)
				return (True, output_path)
			except Exception as error:
				return (False, str(error))
		return (False, '')

	def read_model(self, model_path):
		try:
			model = MetabolicModel(CobraMetabolicModel(model_path))
			model_id = model.id()
			reactions = len(model.reactions())
			metabolites = len(model.metabolites())
			genes = len(model.genes())
			reactions_list = [x.id for x in model.reactions()]
			return (True, None, model, model_id, reactions, metabolites, genes, reactions_list)
		except Exception as error:
			print(error)
			return (False, str(error), None, None, None, None, None, None)

	def print_something(self):
		pass



from findCPcore.AbstractMetabolicModel import AbstractMetabolicModel


class MetabolicModel(AbstractMetabolicModel):
	""" Class that works as an abstraction of a MetabolicModel model

		Attributes
			__model (): object that implements MetabolicModel
	"""

	__model = None

	def __init__(self, model):
		self.__model = model

	def read_model(self, path):
		self.__model.read_model(path)

	def save_model(self, path):
		self.__model.save_model(path)

	def set_state(self, key):
		self.__model.set_state(key)

	def get_state(self, key):
		return self.__model.get_state(key)

	def model(self):
		return self.__model.model()

	def id(self):
		return self.__model.id()

	def compartments(self):
		return self.__model.compartments()

	def objective(self):
		return self.__model.objective()

	def set_objective(self, reaction):
		self.__model.set_objective(reaction)

	def objective_value(self):
		return self.__model.objective_value()

	def reactions(self):
		return self.__model.reactions()

	def metabolites(self):
		return self.__model.metabolites()

	def dem(self):
		return self.__model.dem()

	def chokepoints(self):
		return self.__model.chokepoints()

	def get_fva(self):
		return self.__model.get_fva()

	def genes(self):
		return self.__model.genes()

	def essential_genes(self):
		return self.__model.essential_genes()

	def essential_genes_reactions(self):
		return self.__model.essential_genes_reactions()

	def essential_reactions(self):
		return self.__model.essential_reactions()

	def find_dem(self, compartment="ALL"):
		return self.__model.find_dem(compartment)

	def find_chokepoints(self, exclude_dead_reactions=False):
		return self.__model.find_chokepoints(exclude_dead_reactions)
	
	def remove_dem(self, delete_exchange=False, keep_all_incomplete_reactions=True):
		self.__model.remove_dem(delete_exchange, keep_all_incomplete_reactions)

	def fva(self, loopless=False, verbose=False, update_flux=False, threshold=None, pfba_factor=None):
		return self.__model.fva(loopless, verbose, update_flux, threshold, pfba_factor)

	def find_essential_genes_1(self):
		return self.__model.find_essential_genes_1()

	def find_essential_genes_reactions(self):
		return self.__model.find_essential_genes_reactions()

	def get_growth(self):
		return self.__model.get_growth()

	def find_essential_reactions_1(self):
		self.__model.find_essential_reactions_1()

	def print_model_info(self):
		self.__model.print_model_info()

	def print_metabolites(self, ordered=False):
		self.__model.print_metabolites(ordered)

	def print_reactions(self, ordered=False):
		self.__model.print_reactions(ordered)

	def print_genes(self, ordered=False):
		self.__model.print_genes(ordered)

	def print_dem(self, ordered=False, compartment="ALL"):
		self.__model.print_dem(ordered, compartment)

	def print_chokepoints(self, ordered=False):
		self.__model.print_chokepoints(ordered)

	def print_essential_genes(self, ordered=False):
		self.__model.print_essential_genes(ordered)

	def print_essential_reactions(self, ordered=False):
		self.__model.print_essential_reactions(ordered)

	def print_essential_genes_reactions(self, ordered=False):
		self.__model.print_essential_genes_reactions(ordered)



from abc import ABC, abstractmethod


class AbstractMetabolicModel(ABC):
	""" Abstract class defining the interface of a metabolic network

	"""

	@abstractmethod
	def read_model(self, path):
		""" Loads a metabolic network from a file

		Args:
			path (): Direction containing the metabolic network
		"""
		pass

	@abstractmethod
	def save_model(self, path):
		""" Saves metabolic network to a file

        Args:
            path (): File path
        """
		pass


	@abstractmethod
	def set_state(self, key):
		pass

	@abstractmethod
	def get_state(self, key):
		pass

	@abstractmethod
	def model(self):
		pass

	@abstractmethod
	def id(self):
		pass

	@abstractmethod
	def objective(self):
		""" Returns the objective function of the model

		"""
		pass

	@abstractmethod
	def set_objective(self, reaction):
		""" Sets the reaction param as the objective function of the model

		:param reaction: reaction to set as objective
		"""
		pass

	@abstractmethod
	def objective_value(self):
		pass

	@abstractmethod
	def compartments(self):
		pass

	@abstractmethod
	def reactions(self):
		pass

	@abstractmethod
	def metabolites(self):
		pass

	@abstractmethod
	def dem(self):
		pass

	@abstractmethod
	def chokepoints(self):
		pass

	@abstractmethod
	def get_fva(self):
		pass

	@abstractmethod
	def essential_genes(self):
		pass

	@abstractmethod
	def essential_genes_reactions(self):
		pass

	@abstractmethod
	def essential_reactions(self):
		pass


	@abstractmethod	
	def find_dem(self, compartment):
		""" Finds the dead end metabolites of the metabolic network in a selected compartment

		Args:
			compartment (): Compartment in which the search is done.
		"""
		pass


	@abstractmethod
	def find_chokepoints(self, exclude_dead_reactions):
		""" Finds the chokepoint reactions of the metabolic network

		Args:
			loopless (): Exclude dead reactions from the computed list of chokepoints

		"""
		pass


	@abstractmethod	
	def remove_dem(self, remove_exchange):
		""" While not changing removes the dead end metabolites and reactions only consuming/producing
			of the network

		"""
		pass


	@abstractmethod
	def fva(self, loopless, verbose, update_flux, threshold, pfba_factor):
		""" If possible, runs a Flux Variability Analysis (F.V.A) on the model.

		Args:
			loopless (): Avoids loops of the standard F.V.A.
			verbose (): Print the reusult of the f.v.a.
			update_flux (): Update the upper and lower bounds of the reactions with the ones obtained
							with the f.v.a.
		"""
		pass


	@abstractmethod
	def get_growth(self):
		""" Runs Flux Balance Analysis on the model and returns the objective value.

		"""
		pass


	@abstractmethod
	def find_essential_genes_1(self):
		""" Finds the essential genes of the metabolic network

		"""
		pass


	@abstractmethod
	def find_essential_genes_reactions(self):
		""" Finds the reactions associated to essential genes

		"""
		pass

	@abstractmethod
	def find_essential_reactions_1(self):
		pass


	@abstractmethod
	def print_model_info(self):
		""" Shows info about the metabolic network

		"""
		pass


	@abstractmethod
	def print_metabolites(self, ordered):
		""" Shows the metabolites of the metabolic network

			Args:
				ordered (): Print the metabolites in alphabetic order
		"""
		pass


	@abstractmethod
	def print_reactions(self, ordered):
		""" Shows the reactions of the metabolic network

			Args:
				ordered (): Print the reactions in alphabetic order
		"""
		pass


	@abstractmethod
	def print_genes(self, sorted):
		""" If possible, shows the essential genes of the metabolic network

		Args:
			ordered (): Print the genes in alphabetic order
		"""
		pass

	@abstractmethod
	def print_dem(self, ordered, compartment):
		""" Shows the dead end metabolites of the metabolic network

			Args:
				ordered (): Print the metabolites in alphabetic order
				compartment (): Restric the dead end metabolites shown to a given compartment
		"""
		pass


	@abstractmethod
	def print_chokepoints(self, ordered):
		""" Shows the chokepoint reactions of the metabolic network

			Args:
				ordered (): Print the chokepoint reactions in alphabetic order
		"""
		pass


	@abstractmethod
	def print_essential_genes(self, sorted):
		""" If possible, shows the essential genes of the metabolic network

		Args:
			ordered (): Print the genes in alphabetic order
		"""
		pass

	@abstractmethod
	def print_essential_reactions(self, sorted):
		""" If possible, shows the essential reactions of the metabolic network

		Args:
			ordered (): Print the reactions in alphabetic order
		"""
		pass

	@abstractmethod
	def print_essential_genes_reactions(self, ordered):
		""" If possible, shows the reactions associated to the essential genes of the metabolic network

		Args:
			ordered (): Print the reactions in alphabetic order
		"""
		pass



class Reaction:
	id = None
	name = None
	reaction = None
	upper_bound = None
	lower_bound = None
	gene_reaction_rule = None
	metabolites = None
	genes = None

	def __init__(self, id, name, reaction, upper_bound, lower_bound, gene_reaction_rule):
		self.id = id
		self.name = name
		self.reaction = reaction
		self.upper_bound = upper_bound
		self.lower_bound = lower_bound
		self.gene_reaction_rule = gene_reaction_rule
		self.metabolites = []
		self.genes = []

	def add_metabolite(self, metabolite):
		self.metabolites.append(metabolite)

	def add_gene(self, genes):
		self.genes.append(genes)

class Metabolite:
	id = None
	name = None
	compartment = None
	reactions = None

	def __init__(self, id, name, compartment):
		self.id = id
		self.name = name
		self.compartment = compartment
		self.reactions = []

	def add_reaction(self, reaction):
		self.reactions.append(reaction)

class Gene:
	id = None
	name = None
	reactions = None

	def __init__(self, id, name):
		self.id = id
		self.name = name
		self.reactions = []

	def add_reaction(self, reaction):
		self.reactions.append(reaction)

class State:
	__id = None
	__objective = None
	__objective_value = None

	__reactions = None
	__metabolites = None
	__genes = None

	__dem = None
	__chokepoints = None
	__fva = None
	__essential_genes = None
	__essential_genes_reactions = None
	__essential_reactions = None

	def id(self):
		return self.__id

	def set_id(self, id):
		self.__id = id

	def objective(self):
		return self.__objective

	def set_objective(self, objective):
		self.__objective = objective

	def objective_value(self):
		return self.__objective_value

	def set_objective_value(self, objective_value):
		self.__objective_value = objective_value

	def reactions(self):
		return self.__reactions

	def set_reactions(self, reactions):
		self.__reactions = reactions

	def metabolites(self):
		return self.__metabolites

	def set_metabolites(self, metabolites):
		self.__metabolites = metabolites

	def genes(self):
		return self.__genes

	def set_genes(self, genes):
		self.__genes = genes

	def dem(self):
		return self.__dem

	def set_dem(self, dem):
		self.__dem = dem

	def chokepoints(self):
		return self.__chokepoints

	def set_chokepoints(self, chokepoints):
		self.__chokepoints = chokepoints

	def fva(self):
		return self.__fva

	def set_fva(self, fva):
		self.__fva = fva

	def essential_genes(self):
		return self.__essential_genes

	def set_essential_genes(self, essential_genes):
		self.__essential_genes = essential_genes

	def essential_genes_reactions(self):
		return self.__essential_genes_reactions

	def set_essential_genes_reactions(self, essential_genes_reactions):
		self.__essential_genes_reactions = essential_genes_reactions

	def essential_reactions(self):
		return self.__essential_reactions

	def set_essential_reactions(self, essential_reactions):
		self.__essential_reactions = essential_reactions

	def reversible_reactions(self):
		return self.__reversible_reactions

	def set_reversible_reactions(self, reversible_reactions):
		self.__reversible_reactions = reversible_reactions

	def dead_reactions(self):
		return self.__dead_reactions

	def set_dead_reactions(self, dead_reactions):
		self.__dead_reactions = dead_reactions


class CobraMetabolicStateBuilder:

	reactions = None
	metabolites = None
	genes = None

	def buildState(self, model):
		state = State()
		state.set_id(self.id(model))
		state.set_objective(self.objective(model))
		state.set_objective_value(self.objective_value(model))
		state.set_reversible_reactions(self.reversible_reactions(model))
		state.set_dead_reactions(self.dead_reactions(model))
		(metabolites, reactions, genes) = self.metabolic_network(model)
		state.set_reactions(reactions)
		state.set_metabolites(metabolites)
		state.set_genes(genes)
		try:
			state.set_dem(self.dem(model))
		except Exception:
			state.set_dem({})
		try:
			state.set_chokepoints(self.chokepoints(model))
		except Exception:
			state.set_chokepoints([])
		try:
			state.set_fva(self.fva(model))
		except Exception:
			state.set_fva([])
		try:
			state.set_essential_genes(self.essential_genes(model))
		except Exception:
			state.set_essential_genes([])
		try:
			state.set_essential_genes_reactions(self.essential_genes_reactions(model))
		except Exception:
			state.set_essential_genes_reactions({})
		try:
			state.set_essential_reactions(self.essential_reactions(model))
		except Exception:
			state.set_essential_reactions({})
		return state

	def id(self, model):
		return model.id()

	def objective(self, model):
		return model.objective()

	def objective_value(self, model):
		if model.objective_value() is None:
			return None
		else:
			return model.objective_value()

	def metabolic_network(self, model):
		metabolites = {}
		for metabolite in model.metabolites():
			metabolites[metabolite.id] = Metabolite(metabolite.id, metabolite.name, metabolite.compartment)
		reactions = {}
		for reaction in model.reactions():
			reactions[reaction.id] = Reaction(reaction.id, reaction.name, reaction.reaction, reaction.upper_bound, reaction.lower_bound, reaction.gene_reaction_rule)
		genes = {}
		for gene in model.genes():
			genes[gene.id] = Gene(gene.id, gene.name)

		self.metabolites = metabolites
		self.reactions = reactions
		self.genes = genes

		for reaction in model.reactions():
			new_reaction = reactions[reaction.id]
			for gene in reaction.genes:
				new_gene = genes[gene.id]
				new_reaction.add_gene(new_gene)
		for gene in model.genes():
			new_gene = genes[gene.id]
			for reaction in gene.reactions:
				new_reaction = reactions[reaction.id]
				new_gene.add_reaction(new_reaction)
		for metabolite in model.metabolites():
			new_metabolite = metabolites[metabolite.id]
			for reaction in metabolite.reactions:
				new_reaction = reactions[reaction.id]
				new_metabolite.add_reaction(new_reaction)

		return (list(metabolites.values()), list(reactions.values()), list(genes.values()))


	def dem(self, model):
		result = None
		dem = model.dem()
		if dem is not None:
			result = {}
			for compartment in dem.keys():
				list_c = [self.metabolites[metabolite.id] for metabolite in dem[compartment]]
				result[compartment] = list_c
		return result

	def chokepoints(self, model):
		result = None
		chokepoints = model.chokepoints()
		if chokepoints is not None:
			result = []
			for (reaction, metabolite) in chokepoints:
				result.append((self.reactions[reaction.id], self.metabolites[metabolite.id]))
		return result

	def fva(self, model):
		try:
			result = []
			model_fva = model.get_fva()
			if model_fva is not None:
				for el in model_fva:
					(reaction, upper, lower) = el
					local_reaction = self.reactions[reaction.id]
					result.append((local_reaction, upper, lower))
			return result
		except Exception:
			return []

	def essential_genes(self, model):
		try:
			result = []
			genes = model.essential_genes()
			if genes is not None:
				gen_list = [self.genes[gen.id] for gen in genes]
				result = gen_list
			return result
		except Exception:
			return []

	def essential_genes_reactions(self, model):
		try:
			result = {}
			egr = model.essential_genes_reactions()
			if egr is not None:
				result = {}
				for reaction in egr:
					for gen in egr[reaction]:
						if reaction.id not in result:
							result[self.reactions[reaction.id]] = [self.genes[gen.id]]
						else:
							result[self.reactions[reaction.id]].append(self.genes[gen.id])
			return result
		except Exception:
			return {}

	def essential_reactions(self, model):
		try:
			result = None
			essr = model.essential_reactions()
			if essr is not None:
				result = {}
				for reaction in essr.keys():
					result[self.reactions[reaction.id]] = essr[reaction]
			return result
		except Exception:
			return {}

	def reversible_reactions(self, model):
		return [r.id for r in model.reversible_reactions()]

	def dead_reactions(self, model):
		return [r.id for r in model.dead_reactions()]



import threading
import inspect
import ctypes
import time

from findCPcore.FacadeUtils import FacadeUtils

TASK_READ_MODEL = "READ_MODEL"
TASK_SAVE_DEM = "SAVE_DEM"
TASK_SAVE_FVA = "SAVE_FVA"
TASK_SAVE_FVA_DEM = "SAVE_FVA_DEM"
TASK_SPREADSHEET = "SPREADSHEET"
TASK_SENSIBILITY = "TASK_SENSIBILITY"
TASK_SAVE_SPREADSHEET = "SAVE_SPREADSHEET"
TASK_FIND_AND_REMOVE_DEM = "TASK_FIND_AND_REMOVE_DEM"
TASK_SAVE_MODEL = "TASK_SAVE_MODEL"
TASK_FVA = "TASK_FVA"


class ThreadInterrupt(Exception):
	def __str__(self):
		return "Exception: Thread stopped"

	def _async_raise(tid, exctype, syserr=True):
		'''Raises an exception in the threads with id tid'''
		if not inspect.isclass(exctype):
			raise TypeError("Only types can be raised (not instances)")
		res = ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(tid),
	                                                     ctypes.py_object(exctype))
		if res == 0:
			raise ValueError("invalid thread id")
		elif res != 1:
			# "if it returns a number greater than one, you're in trouble,
			# and you should call it again with exc=NULL to revert the effect"
			ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(tid), None)
		if syserr:
			raise SystemError("PyThreadState_SetAsyncExc failed")


class FacadeThread(threading.Thread):

	def get_my_tid(self):
		"""determines this (self's) thread id

		CAREFUL : this function is executed in the context of the caller
		thread, to get the identity of the thread represented by this
		instance.
		"""
		if not self.isAlive():
			raise threading.ThreadError("the thread is not active")

		# do we have it cached?
		if hasattr(self, "_thread_id"):
			return self._thread_id

		# no, look for it in the _active dict
		for tid, tobj in threading._active.items():
			if tobj is self:
				self._thread_id = tid
				return tid

		raise AssertionError("could not determine the thread's id")

	def raiseExc(self, exctype, tid=None):
		"""Raises the given exception type in the context of this thread.

		If the thread is busy in a system call (time.sleep(),
		socket.accept(), ...), the exception is simply ignored.

		If you are sure that your exception should terminate the thread,
		one way to ensure that it works is:

			t = ThreadWithExc( ... )
			...
			t.raiseExc( SomeException )
			while t.isAlive():
				time.sleep( 0.1 )
				t.raiseExc( SomeException )

		If the exception is to be caught by the thread, you need a way to
		check that your thread has caught it.

		CAREFUL : this function is executed in the context of the
		caller thread, to raise an excpetion in the context of the
		thread represented by this instance.
		"""
		if tid is None:
			_async_raise(self.get_my_tid(), exctype)
		else:
			_async_raise(tid, exctype)

	def __init__(self, model_path):
		self.model_path = model_path
		super(FacadeThread, self).__init__()

	def set_task(self, task, notify_function=None, args1=None, args2=None, output_path=None, objective=None, fraction=1.0):
		self.task = task
		self.notify_function = notify_function
		self.args1=args1
		self.args2=args2
		self.output_path = output_path
		self.objective = objective
		self.fraction = fraction


	def run(self):
		try:
			if self.task == TASK_READ_MODEL:
				f = FacadeUtils()
				(result, error, model, model_id, reactions, metabolites, genes, reactions_list) = f.read_model(self.model_path)
				self.notify_function(result, error,  model_id, reactions, metabolites, genes, reactions_list, self.args1, self.args2)

			elif self.task == TASK_FIND_AND_REMOVE_DEM:
				f = FacadeUtils()
				self.model = f.find_and_remove_dem(self.model_path)
				if self.output_path is not None and self.model is not None:
					f.save_model(self.output_path, self.model)
				self.notify_function("text", self.args1, self.args2, ended=True, result=True, error=None)

			elif self.task == TASK_SAVE_FVA:
				f = FacadeUtils()
				(self.model, error) = f.run_fva(self.model_path, objective=self.objective, fraction=self.fraction)
				result_ok = error == ""
				if result_ok:
					if self.output_path is not None and self.model is not None:
						f.save_model(self.output_path, self.model)
					self.notify_function("text", self.args1, self.args2, ended=True, result=True, error=None)
				else:
					self.notify_function("text", self.args1, self.args2, ended=True, result=False, error=error)

			elif self.task == TASK_SAVE_FVA_DEM:
				f = FacadeUtils()
				(self.model, error) = f.run_fva_remove_dem(self.model_path, objective=self.objective, fraction=self.fraction)
				result_ok = error == ""
				if result_ok:
					if self.output_path is not None and self.model is not None:
						f.save_model(self.output_path, self.model)
					self.notify_function("text", self.args1, self.args2, ended=True, result=True, error=None)
				else:
					self.notify_function("text", self.args1, self.args2, ended=True, result=False, error=error)

			elif self.task == TASK_SPREADSHEET:
				f = FacadeUtils()
				s = f.run_summary_model(self.model_path, self.notify_function, self.args1, None, objective=self.objective, fraction=self.fraction)
				self.spreadsheet = s
				if self.output_path is not None:
					(result_ok, error) = f.save_spreadsheet(self.output_path, s)
				self.notify_function("text", self.args1, self.args2, ended=True, result=True, error=None)

			elif self.task == TASK_SENSIBILITY:
				f = FacadeUtils()
				s = f.run_sensibility_analysis(self.model_path, self.notify_function, self.args1, None, objective=self.objective)
				self.spreadsheet = s
				if self.output_path is not None:
					(result_ok, error) = f.save_spreadsheet(self.output_path, s)
				self.notify_function("text", self.args1, self.args2, ended=True, result=True, error=None)

			elif self.task == TASK_SAVE_SPREADSHEET:
				f = FacadeUtils()
				s = self.spreadsheet
				if self.output_path is not None and s is not None:
					(result_ok, error) = f.save_spreadsheet(self.output_path, s)
					self.notify_function("text", self.args1, self.args2, ended=True, result=result_ok, error=error)

			elif self.task == TASK_SAVE_MODEL:
				f = FacadeUtils()
				(result, text) = f.save_model(self.output_path, self.model)
				self.notify_function("text", self.args1, self.args2, ended=True, result=result, error=text)
		except Exception as error:
			# Thread stopped
			# This raise is just for debugging purposes
			print("DEBUG: please remove in FacadeThread.py:", str(error))
			raise error
			print("Error:", str(error))
			pass




import threading
import inspect
import ctypes
import time

from findCPcore.FacadeUtils import FacadeUtils
from findCPcore.FacadeThread import FacadeThread, ThreadInterrupt

TASK_READ_MODEL = "READ_MODEL"
TASK_SAVE_DEM = "SAVE_DEM"
TASK_SAVE_FVA = "SAVE_FVA"
TASK_SAVE_FVA_DEM = "SAVE_FVA_DEM"
TASK_SPREADSHEET = "SPREADSHEET"
TASK_SENSIBILITY = "TASK_SENSIBILITY"
TASK_SAVE_SPREADSHEET = "SAVE_SPREADSHEET"
TASK_FIND_AND_REMOVE_DEM = "TASK_FIND_AND_REMOVE_DEM"
TASK_SAVE_MODEL = "TASK_SAVE_MODEL"
TASK_FVA = "TASK_FVA"

class Facade:

	model_path = None
	model = None
	spreadsheet = None

	model_id = None
	reactions = None
	metabolites = None
	genes = None

	thread1 = None
	tid = None

	def isAlive_tid(self, tid_input):
		for tid, tobj in threading._active.items():
			if tid == tid_input:
				return True
		return False

	def stop_tid(self, tid):
		try:
			ThreadInterrupt._async_raise(tid, ThreadInterrupt, syserr=False)
			print("STOPPING")
			while self.isAlive_tid(tid):
				time.sleep(0.1)
				print("STOPPING")
				ThreadInterrupt._async_raise(tid, ThreadInterrupt, syserr=False)
		except Exception as errr:
			# Thread stopped - catch some derivated errors
			#print("DEBUG: PLEASE REMOVE THIS RAISE IN Facade.py")
			#raise errr
			pass

	def get_tid(self):
		return self.thread1.get_my_tid()

	def stop(self):
		try:
			tid = self.get_tid()			
			self.stop_tid(tid)
		except Exception as exc:
			# Thread is already stopped
			#print("DEBUG: PLEASE REMOVE THIS RAISE IN Facade.py")
			#raise exc			
			pass

	def set_model_path(self, model_path):
		self.model_path = model_path

	def read_model(self, stoppable, model_path, notify_function, args1=None, args2=None):
		self.model_path = model_path
		if not stoppable:
			(result, error, model, model_id, reactions, metabolites, genes) = FacadeUtils.read_model(model_path)
			self.model = model
			self.model_id = model_id
			self.reactions = reactions
			self.metabolites = metabolites
			self.genes = genes
			return (result, error)
		else:
			self.thread1 = FacadeThread(self.model_path)
			self.thread1.set_task(TASK_READ_MODEL, notify_function, args1, args2)
			self.thread1.start()
			self.tid = self.thread1.get_my_tid()
		return (None, None)


	def get_model_info(self):
		return (self.model_id, self.reactions, self.metabolites, self.genes)

	def generate_spreadsheet(self, stoppable, model_path, print_f, args1=None, args2=None, output_path=None, objective=None, fraction=1.0):
		if not stoppable:
			f = FacadeUtils()
			self.spreadsheet = f.run_summary_model(model_path, print_f, args1, args2, objective, fraction)
			if self.spreadsheet is not None and output_path is not None:
				self.save_spreadsheet(stoppable, output_path, print_f)
			return ""
		else:
			self.thread1 = FacadeThread(self.model_path)
			self.thread1.set_task(TASK_SPREADSHEET, print_f, args1, args2, output_path, objective=objective, fraction=fraction)
			self.thread1.start()
			self.tid = self.thread1.get_my_tid()

	def generate_sensibility_spreadsheet(self, stoppable, model_path, print_f, args1=None, args2=None, output_path=None, objective=None):
		if not stoppable:
			f = FacadeUtils()
			self.spreadsheet = f.run_sensibility_analysis(model_path, print_f, args1, args2, objective)
			if self.spreadsheet is not None and output_path is not None:
				self.save_spreadsheet(stoppable, output_path, print_f)
			return ""
		else:
			self.thread1 = FacadeThread(self.model_path)
			self.thread1.set_task(TASK_SENSIBILITY, print_f, args1, args2, output_path, objective=objective)
			self.thread1.start()
			self.tid = self.thread1.get_my_tid()


	def save_spreadsheet(self, stoppable, output_path, print_f):
		if not stoppable:
			f = FacadeUtils()
			(result_ok, text) = f.save_spreadsheet(output_path, self.spreadsheet)
			return (result_ok, text)
		else:
			self.thread1.set_task(TASK_SAVE_SPREADSHEET, notify_function=print_f, args1=None, args2=None, output_path=output_path)
			self.thread1.run()
		return (None, None)


	# print_f (arg1, ended=True, str)
	def find_and_remove_dem(self, stoppable, output_path, print_f, model_path, args1=None, args2=None):
		if not stoppable:
			f = FacadeUtils()
			self.model = f.find_and_remove_dem(model_path)
		else:
			self.thread1 = FacadeThread(self.model_path)
			self.thread1.set_task(TASK_FIND_AND_REMOVE_DEM, notify_function=print_f, args1=args1, args2=None,
			                      output_path=output_path)
			self.thread1.start()
			self.tid = self.thread1.get_my_tid()

	def run_fva(self, stoppable, output_path, print_f, model_path, args1=None, args2=None, objective=None, fraction=1.0):
		if not stoppable:
			f = FacadeUtils()
			(self.model, errors) = f.run_fva(model_path, objective, fraction)
			return errors
		else:
			self.thread1 = FacadeThread(self.model_path)
			self.thread1.set_task(TASK_SAVE_FVA, notify_function=print_f, args1=args1, args2=None,
			                      output_path=output_path, objective=objective, fraction=fraction)
			self.thread1.start()
			self.tid = self.thread1.get_my_tid()

	def run_fva_remove_dem(self, stoppable, output_path, print_f, model_path, args1=None, args2=None, objective=None, fraction=1.0):
		if not stoppable:
			f = FacadeUtils()
			(self.model, errors) = f.run_fva_remove_dem(model_path, objective, fraction)
			return errors
		else:
			self.thread1 = FacadeThread(self.model_path)
			self.thread1.set_task(TASK_SAVE_FVA_DEM, notify_function=print_f, args1=args1, args2=None,
			                      output_path=output_path, objective=objective, fraction=fraction)
			self.thread1.start()
			self.tid = self.thread1.get_my_tid()

	# guarda el modelo calculado que se almacena en self
	# print_f (arg, result_ok=True/False, error text)
	def save_model(self, output_path, stoppable, print_f):
		if stoppable:
			self.thread1.set_task(TASK_SAVE_MODEL, notify_function=print_f, args1=None, args2=None,
			                      output_path=output_path)
			self.thread1.run()
		else:
			f = FacadeUtils()
			(result, text) = f.save_model(output_path, self.model)
			return (result, text)







import xlwt
import itertools
from math import isnan

class Spreadsheet:

	__spreadsheet = None

	
	def spreadsheet_init(self):
		""" Initialize the xlwt Workbook

		"""
		self.__spreadsheet = xlwt.Workbook()

	def set_workbook(self, workbook):
		self.__spreadsheet = workbook
		

	def spreadsheet_write_model_info(self, state, sheet_name):
		""" Writes in a sheet of the xlwt Workbook '__spreadsheet' info of the model with the following format:

				MODEL_ID | REACTIONS | METABOLITES | GENES
				---------+-----------+-------------+-------
				model id | len(reac.)| len(metab.) | len(genes)

		Args:
			sheet_name (): Name of the new sheet. Two diffetent sheets can't have the same name.
		"""
		if self.__spreadsheet is None:
			self.spreadsheet_init()

		style = xlwt.XFStyle()
		font = xlwt.Font()
		font.bold = True
		style.font = font
		sheet = self.__spreadsheet.add_sheet(sheet_name)
		sheet.write(0, 0, "MODEL_ID", style=style)
		sheet.write(0, 1, state.id())
		sheet.write(1, 0, "REACTIONS", style=style)
		sheet.write(1, 1, len(state.reactions()))
		sheet.write(2, 0, "METABOLITES", style=style)
		sheet.write(2, 1, len(state.metabolites()))
		sheet.write(3, 0, "GENES", style=style)
		sheet.write(3, 1, len(state.genes()))
		sheet.write(4, 0, "OBJECTIVE FUNCTION", style=style)
		sheet.write(4, 1, state.objective())


	def __id(self, e):
		return e.id



	def __spreadsheet_write_metabolites(self, list, sheet_name, ordered=True, print_reactions=False):
		""" Writes in a sheet of the xlwt Workbook '__spreadsheet' the metabolites of the model with the following format:

			METABOLITE ID | METABOLITE NAME | METABOLITE COMPARTMENT | REACTION ID | REACTION NAME | REACTION
			--------------+-----------------+------------------------+-------------+---------------+----------
			mtb 1         | metabolite 1    | compartment 1          | react1      | reaction 1    | mtb 1 -> mtb 2
			--------------+-----------------+------------------------+-------------+---------------+----------
						  |                 |                        | react2      | reaction 2    | mtb 1 -> mtb 3

		Args:
			list (): List of cobra.core.metabolites to be written.
			sheet_name (): Name of the new sheet. Two diffetent sheets can't have the same name.
			ordered (): Write the metabolites in alphabetical order by id.
			print_reactions (): Write the reactions in which a metabolite appears.
		"""
		if self.__spreadsheet is None:
			self.spreadsheet_init()
		metabolites = list
		if ordered:
			metabolites.sort(key=self.__id)

		style = xlwt.XFStyle()
		font = xlwt.Font()
		font.bold = True
		style.font = font
		sheet = self.__spreadsheet.add_sheet(sheet_name)
		sheet.write(0, 0, "METABOLITE ID", style=style)
		sheet.write(0, 1, "METABOLITE NAME", style=style)
		sheet.write(0, 2, "METABOLITE COMPARTMENT", style=style)
		if print_reactions:
			sheet.write(0, 3, "REACTION ID", style=style)
			sheet.write(0, 4, "REACTION NAME", style=style)
			sheet.write(0, 5, "UPPER BOUND", style=style)
			sheet.write(0, 6, "LOWER BOUND", style=style)
			sheet.write(0, 7, "REACTION", style=style)
		i = 1
		for metabolite in metabolites:
			sheet.write(i, 0, metabolite.id)
			sheet.write(i, 1, metabolite.name)
			sheet.write(i, 2, metabolite.compartment)
			if print_reactions:
				reactions = []
				for r in metabolite.reactions:
					reactions.append(r)
				reactions.sort(key=self.__id)
				for reaction in reactions:
					sheet.write(i, 3, reaction.id)
					sheet.write(i, 4, reaction.name)
					sheet.write(i, 5, reaction.upper_bound)
					sheet.write(i, 6, reaction.lower_bound)
					sheet.write(i, 7, reaction.reaction)
					i = i + 1
			if not print_reactions or len(reactions) == 0:
				i = i + 1



	def spreadsheet_write_reactions(self, state, sheet_name, ordered=True, print_genes=False):
		"""  Writes in a sheet of the xlwt Workbook '__spreadsheet' the reactions of the model with the following format:

			REACTION ID | REACTION NAME | REACTION | UPPER BOUND | LOWER BOUND | CHOKEPOINT | ESSENTIAL GENE
			------------+---------------+----------+-------------+-------------+------------+----------------
			react 1     | reaction 1    | mt1->mt2 | 1000        | 0           | TRUE       |
			------------+---------------+----------+-------------+-------------+------------+----------------
			react 2     | reaction 2    | mt3->mt4 | 2000        | 0           |            | TRUE

		Args:
			sheet_name (): Name of the new sheet. Two diffetent sheets can't have the same name.
			ordered (): Print the reactions in alphabetical order by id
			tag_chokepoints (): mark a reaction with 'TRUE' if chokepoint reaction
			tag_essential_genes (): mark a reactions with 'TRUE' if associated to an essential gene
		"""
		if self.__spreadsheet is None:
			self.spreadsheet_init()

		reactions = state.reactions()
		if ordered:
			reactions.sort(key=self.__id)

		style = xlwt.XFStyle()
		font = xlwt.Font()
		font.bold = True
		style.font = font
		sheet = self.__spreadsheet.add_sheet(sheet_name)
		sheet.write(0, 0, "REACTION ID", style=style)
		sheet.write(0, 1, "REACTION NAME", style=style)
		sheet.write(0, 2, "REACTION", style=style)
		sheet.write(0, 3, "UPPER BOUND", style=style)
		sheet.write(0, 4, "LOWER BOUND", style=style)
		if print_genes:
			sheet.write(0, 5, "GENE RELATION", style=style)
			sheet.write(0, 6, "GENE ID", style=style)
		i = 1
		for reaction in reactions:
			sheet.write(i, 0, reaction.id)
			sheet.write(i, 1, reaction.name)
			sheet.write(i, 2, reaction.reaction)
			sheet.write(i, 3, reaction.upper_bound)
			sheet.write(i, 4, reaction.lower_bound)
			if print_genes:
				sheet.write(i, 5, reaction.gene_reaction_rule)
				for gen in reaction.genes:
					sheet.write(i, 6, gen.id)
					i = i + 1
			if not print_genes or len(reaction.genes) == 0:
				i = i + 1


	def spreadsheet_write_metabolites(self, state, sheet_name, ordered=True, print_reactions=False):
		""" Writes the metabolites of the model with the function '__spreadsheet_write_metabolites'

		"""
		self.__spreadsheet_write_metabolites(state.metabolites(), sheet_name, ordered, print_reactions)


	def spreadsheet_write_genes(self, state, sheet_name, ordered=False, print_reactions=False):
		if self.__spreadsheet is None:
			self.spreadsheet_init()

		genes = state.genes()
		if ordered:
			genes.sort(key=self.__id)

		style = xlwt.XFStyle()
		font = xlwt.Font()
		font.bold = True
		style.font = font
		sheet = self.__spreadsheet.add_sheet(sheet_name)
		sheet.write(0, 0, "GENE ID", style=style)
		sheet.write(0, 1, "GENE NAME", style=style)
		if print_reactions:
			sheet.write(0, 2, "GENE RELATION", style=style)
			sheet.write(0, 3, "REACTION ID", style=style)
			sheet.write(0, 4, "REACTION NAME", style=style)
		i = 1
		for gen in genes:
			sheet.write(i, 0, gen.id)
			sheet.write(i, 1, gen.name)
			if print_reactions:
				reactions = []
				for r in gen.reactions:
					reactions.append(r)
				reactions.sort(key=self.__id)
				for reaction in reactions:
					sheet.write(i, 2, reaction.gene_reaction_rule)
					sheet.write(i, 3, reaction.id)
					sheet.write(i, 4, reaction.name)
					i = i + 1
			if not print_reactions or len(reactions) == 0:
				i = i + 1

	def spreadsheet_write_dem(self, state, sheet_name, ordered=False, print_reactions=False, compartment="ALL"):
		""" Writes the dead end metabolites of the model with the function '__spreadsheet_write_metabolites'

		"""
		if state.dem() is not None:
			if compartment == "ALL":
				all_dems = list(itertools.chain.from_iterable(state.dem().values()))
				self.__spreadsheet_write_metabolites(all_dems, sheet_name, ordered, print_reactions)
			else:
				if compartment in state.dem().keys():
					self.__spreadsheet_write_metabolites(state.dem()[compartment], sheet_name, ordered, print_reactions)


	def __reaction_id(self, obj):
		""" Given an object _MetaboliteReact it returns the 'id' of the reaction

		Args:
			obj (): Object cobra.core.reaction

		Returns: 'id' field of the reaction

		"""
		(reaction, metabolite) = obj
		return reaction.id

	def spreadsheet_write_chokepoints(self, state, sheet_name, ordered=False):
		""" 
		"""
		if self.__spreadsheet is None:
			self.spreadsheet_init()
		if state.chokepoints() is not None:
			chokepoints = state.chokepoints()
			if ordered:
				chokepoints.sort(key=self.__reaction_id)
			style = xlwt.XFStyle()
			font = xlwt.Font()
			font.bold = True
			style.font = font
			sheet = self.__spreadsheet.add_sheet(sheet_name)
			sheet.write(0, 0, "REACTION ID", style=style)
			sheet.write(0, 1, "METABOLITE ID", style=style)
			i = 1
			for (reaction, metabolite) in chokepoints:
				sheet.write(i, 0, reaction.id)
				sheet.write(i, 1, metabolite.id)
				i = i + 1

	def spreadsheet_write_reversible_reactions(self, sheet_name, state_initial, state_fva, ordered=False):
		""" 
		"""
		if self.__spreadsheet is None:
			self.spreadsheet_init()

		reactions = state_initial.reactions()
		if ordered:
			reactions.sort(key=self.__id)

		reversible_initial = state_initial.reversible_reactions()
		reversible_fva = state_fva.reversible_reactions()

		style = xlwt.XFStyle()
		font = xlwt.Font()
		font.bold = True
		style.font = font
		sheet = self.__spreadsheet.add_sheet(sheet_name)
		sheet.write(0, 0, "REACTION ID", style=style)
		sheet.write(0, 1, "REVERSIBLE INITIAL", style=style)
		sheet.write(0, 2, "REVERSIBLE FVA", style=style)
		i = 1			
		for reaction in reactions:
			sheet.write(i, 0, reaction.id)
			if reaction.id in reversible_initial:
				sheet.write(i, 1, "TRUE")
			if reaction.id in reversible_fva:
				sheet.write(i, 2, "TRUE")
			i = i + 1


	def __reaction_id_fva(self, obj):
		(react, max, min) = obj
		return react.id


	def spreadsheet_write_fva(self, state, sheet_name, ordered=False):
		"""  Writes in a sheet of the xlwt Workbook '__spreadsheet' the result of the Flux Variability Analysis.

			REACTION ID | REACTION NAME | REACTION | MAXIMUN | MINIMUM
			------------+---------------+----------+---------+------------
			react 1     | reaction 1    | mt1->mt2 | 1000    | 0
			------------+---------------+----------+---------+------------
			react 2     | reaction 2    | mt3->mt4 | 2000    |

		Args:
			sheet_name (): Name of the new sheet. Two diffetent sheets can't have the same name.
			ordered (): Print the reactions in alphabetical order by id
		"""
		if self.__spreadsheet is None:
			self.spreadsheet_init()

		if state.fva() is not None:
			fva = state.fva()
			if ordered:
				fva.sort(key=self.__reaction_id_fva)
			style = xlwt.XFStyle()
			font = xlwt.Font()
			font.bold = True
			style.font = font
			sheet = self.__spreadsheet.add_sheet(sheet_name)
			sheet.write(0, 0, "REACTION ID", style=style)
			sheet.write(0, 1, "REACTION NAME", style=style)
			sheet.write(0, 2, "MAX", style=style)
			sheet.write(0, 3, "MIN", style=style)
			i = 1
			for obj in fva:
				(reaction, max, min) = obj
				sheet.write(i, 0, reaction.id)
				sheet.write(i, 1, reaction.name)
				sheet.write(i, 2, max)
				sheet.write(i, 3, min)
				i = i + 1


	def spreadsheet_write_essential_genes(self, state, sheet_name, ordered=False):
		"""  If the model has genes, writes in a sheet of the xlwt Workbook '__spreadsheet'
			the essential genes of the model with the following format:

			GENE ID | REACTION ID | REACTION NAME | REACTION
			--------+-------------+---------------+-------------
			gene 1  | react 1     | reaction 1    | mt1 + mt2 -> mt3
			--------+-------------+---------------+-------------
			        | react 2     | reaction 2    | mt4 -> mt5
			--------+-------------+---------------+-------------
			gene 2  | react 1     | reaction 1    | mt1 + mt2 -> mt3

		Args:
			sheet_name (): Name of the new sheet. Two diffetent sheets can't have the same name.
			ordered (): Print the genes in alphabetical order by id
		"""
		if self.__spreadsheet is None:
			self.spreadsheet_init()

		if state.essential_genes() is not None:

			style = xlwt.XFStyle()
			font = xlwt.Font()
			font.bold = True
			style.font = font
			sheet = self.__spreadsheet.add_sheet(sheet_name)
			sheet.write(0, 0, "GENE ID", style=style)
			sheet.write(0, 1, "GENE NAME", style=style)
			sheet.write(0, 2, "REACTION GENE RELATION", style=style)
			sheet.write(0, 3, "REACTION ID", style=style)
			sheet.write(0, 4, "REACTION NAME", style=style)

			genes = state.essential_genes()
			if ordered:
				genes.sort(key=self.__id)
			row = 1
			for gene in genes:
				reactions = gene.reactions
				sheet.write(row, 0, gene.id)
				sheet.write(row, 1, gene.name)
				sheet.write(row, 2, reactions[0].gene_reaction_rule)
				sheet.write(row, 3, reactions[0].id)
				sheet.write(row, 4, reactions[0].name)
				row = row + 1
				i = 1
				while i < len(reactions):
					sheet.write(row, 2, reactions[i].gene_reaction_rule)
					sheet.write(row, 3, reactions[i].id)
					sheet.write(row, 4, reactions[i].name)
					i = i + 1
					row = row + 1


	def spreadsheet_write_essential_genes_reactions(self, state, sheet_name, ordered=False):
		"""  If the model has genes, writes in a sheet of the xlwt Workbook '__spreadsheet'
			the reactions associated to the essential genes of the model with the following format:

			REACTION ID | REACTION NAME | REACTION | GENE ID
			------------+---------------+----------+-------------
			react 1     | reaction 1    | mt1->mt2 | gene1
			------------+---------------+----------+-------------
			            |               |          | gene2
			------------+---------------+----------+-------------
			react 2     | reaction 2    | mt1->mt2 | gene3

		Args:
			sheet_name (): Name of the new sheet. Two diffetent sheets can't have the same name.
			ordered (): Print the reactions in alphabetical order by id
		"""
		if self.__spreadsheet is None:
			self.spreadsheet_init()
		if state.essential_genes_reactions() is not None:
			reactions = state.essential_genes_reactions()

			style = xlwt.XFStyle()
			font = xlwt.Font()
			font.bold = True
			style.font = font
			sheet = self.__spreadsheet.add_sheet(sheet_name)
			sheet.write(0, 0, "REACTION ID", style=style)
			sheet.write(0, 1, "REACTION NAME", style=style)
			sheet.write(0, 2, "GENE REACTION RELATION", style=style)
			sheet.write(0, 3, "GENE ID", style=style)
			sheet.write(0, 4, "GENE NAME", style=style)
			reactions_key = reactions.keys()
			if ordered:
				reactions_key = sorted(reactions.keys(), key=self.__id)
			row = 1
			for reaction in reactions_key:
				genes = reactions[reaction]
				sheet.write(row, 0, reaction.id)
				sheet.write(row, 1, reaction.name)
				sheet.write(row, 2, reaction.gene_reaction_rule)
				sheet.write(row, 3, genes[0].id)
				sheet.write(row, 4, genes[0].name)
				row = row + 1
				i = 1
				while i < len(genes):
					sheet.write(row, 3, genes[i].id)
					sheet.write(row, 4, genes[i].name)
					i = i + 1
					row = row + 1

	def spreadsheet_write_essential_reactions(self, sheet_name, state_initial, state_dem, state_fva, state_fva_dem, ordered=True):
		"""
		"""
		if self.__spreadsheet is None:
			self.spreadsheet_init()

		reactions_initial = {}
		if state_initial.essential_reactions() is not None:
			for reaction, growth in state_initial.essential_reactions().items():
				reactions_initial[reaction.id] = growth
		reactions_dem = {}
		if state_dem.essential_reactions() is not None:
			for reaction, growth in state_dem.essential_reactions().items():
				reactions_dem[reaction.id] = growth
		reactions_fva = {}
		if state_fva.essential_reactions() is not None:
			for reaction, growth in state_fva.essential_reactions().items():
				reactions_fva[reaction.id] = growth
		reactions_fva_dem = {}
		if state_fva_dem.essential_reactions() is not None:
			for reaction, growth in state_fva_dem.essential_reactions().items():
				reactions_fva_dem[reaction.id] = growth

		style = xlwt.XFStyle()
		font = xlwt.Font()
		font.bold = True
		style.font = font
		sheet = self.__spreadsheet.add_sheet(sheet_name)
		sheet.write(0, 0, "REACTION ID", style=style)
		sheet.write(0, 1, "REACTION NAME", style=style)
		sheet.write(0, 2, "OBJECTIVE INITIAL", style=style)
		sheet.write(0, 3, "OBJECTIVE DEM", style=style)
		sheet.write(0, 4, "OBJECTIVE FVA", style=style)
		sheet.write(0, 5, "OBJECTIVE FVA DEM", style=style)
		reactions_initial_key = reactions_initial.keys()
		reactions_dem_key = reactions_dem.keys()
		reactions_fva_key = reactions_fva.keys()
		reactions_fva_dem_key = reactions_fva_dem.keys()
		if ordered:
			reactions = sorted(state_initial.essential_reactions().keys(), key=self.__id)
		row = 1
		for reaction in reactions:
			sheet.write(row, 0, reaction.id)
			sheet.write(row, 1, reaction.name)
			if reaction.id in reactions_initial_key:
				sheet.write(row, 2, reactions_initial[reaction.id])
			if reaction.id in reactions_dem_key:
				sheet.write(row, 3, reactions_dem[reaction.id])
			if reaction.id in reactions_fva_key:
				sheet.write(row, 4, reactions_fva[reaction.id])
			if reaction.id in reactions_fva_dem_key:
				sheet.write(row, 5, reactions_fva_dem[reaction.id])
			row = row + 1


	def spreadsheet_write_summary_reactions(self, sheet_name,  state_initial, state_dem, state_fva, state_fva_dem):
		"""

		Args:
			sheet_name ():

		Returns:

		"""
		if self.__spreadsheet is None:
			self.spreadsheet_init()

		reactions = state_initial.reactions()
		reactions.sort(key=self.__id)

		#
		initial_chokepoints_dict = {}
		if state_initial.chokepoints() is not None:
			for el in state_initial.chokepoints():
				(reaction, metabolite) = el
				reaction = reaction.id
				metabolite = metabolite.id
				if reaction in initial_chokepoints_dict:
					initial_chokepoints_dict[reaction].append(metabolite)
				else:
					initial_chokepoints_dict[reaction] = [metabolite]
		#
		chokepoints_dem_dict = {}
		if state_dem.chokepoints() is not None:
			for el in state_dem.chokepoints():
				(reaction, metabolite) = el
				reaction = reaction.id
				metabolite = metabolite.id
				if reaction in chokepoints_dem_dict:
					chokepoints_dem_dict[reaction].append(metabolite)
				else:
					chokepoints_dem_dict[reaction] = [metabolite]
		#
		chokepoints_fva_dict = {}
		if state_fva.chokepoints() is not None:
			for el in state_fva.chokepoints():
				(reaction, metabolite) = el
				reaction = reaction.id
				metabolite = metabolite.id
				if reaction in chokepoints_fva_dict:
					chokepoints_fva_dict[reaction].append(metabolite)
				else:
					chokepoints_fva_dict[reaction] = [metabolite]
		#
		chokepoints_fva_dem_dict = {}
		if state_fva_dem.chokepoints() is not None:
			for el in state_fva_dem.chokepoints():
				(reaction, metabolite) = el
				reaction = reaction.id
				metabolite = metabolite.id
				if reaction in chokepoints_fva_dem_dict:
					chokepoints_fva_dem_dict[reaction].append(metabolite)
				else:
					chokepoints_fva_dem_dict[reaction] = [metabolite]


		style = xlwt.XFStyle()
		font = xlwt.Font()
		font.bold = True
		style.font = font
		sheet = self.__spreadsheet.add_sheet(sheet_name, cell_overwrite_ok=True)
		sheet.write(0, 0, "REACTION ID", style=style)
		sheet.write(0, 1, "REACTION", style=style)
		sheet.write(0, 2, "CHOKEPOINT INITIAL", style=style)
		sheet.write(0, 3, "CHOKEPOINT AFTER DEM REMOVE", style=style)
		sheet.write(0, 4, "CHOKEPOINT AFTER FVA", style=style)
		sheet.write(0, 5, "CHOKEPOINT AFTER FVA AND DEM REMOVE", style=style)
		i = 1
		for reaction in reactions:
			sheet.write(i, 0, reaction.id)
			sheet.write(i, 1, reaction.reaction)
			j = 0
			if reaction.id in initial_chokepoints_dict:
				for metabolite in initial_chokepoints_dict[reaction.id]:
					sheet.write(i + j, 2, metabolite)
					sheet.write(i + j, 6, " ")
					j = j + 1
			else:
				sheet.write(i + j, 2, " ")
			k = 0
			if reaction.id in chokepoints_dem_dict:
				for metabolite in chokepoints_dem_dict[reaction.id]:
					sheet.write(i + k, 3, metabolite)
					sheet.write(i + k, 6, " ")
					k = k + 1
			else:
				sheet.write(i + k, 3, " ")
			l = 0
			if reaction.id in chokepoints_fva_dict:
				for metabolite in chokepoints_fva_dict[reaction.id]:
					sheet.write(i + l, 4, metabolite)
					sheet.write(i + l, 6, " ")
					l = l + 1
			else:
				sheet.write(i + l, 4, " ")
			m = 0
			if reaction.id in chokepoints_fva_dem_dict:
				for metabolite in chokepoints_fva_dem_dict[reaction.id]:
					sheet.write(i + m, 5, metabolite)
					sheet.write(i + m, 6, " ")
					m = m + 1
			else:
				sheet.write(i + m,5, " ")
			maxim = max(j,k,l,m)
			if maxim == 0:
				i = i + 1
			else:
				i = i + maxim


	def spreadsheet_write_summary_metabolites(self,  sheet_name, state_initial, state_fva):

		if self.__spreadsheet is None:
			self.spreadsheet_init()

		style = xlwt.XFStyle()
		font = xlwt.Font()
		font.bold = True
		style.font = font

		aux_dem_init = []
		if state_initial.dem() is not None:
			aux = list(itertools.chain.from_iterable(state_initial.dem().values()))
			aux_dem_init = [e.id for e in aux]
		aux_dem_fva = []
		if state_fva.dem() is not None and state_fva.fva() is not None:
			aux = list(itertools.chain.from_iterable(state_fva.dem().values()))
			aux_dem_fva = [e.id for e in aux]
		metabolites = state_initial.metabolites()
		metabolites.sort(key=self.__id)
		sheet = self.__spreadsheet.add_sheet(sheet_name)
		sheet.write(0, 0, "METABOLITE ID", style=style)
		sheet.write(0, 1, "METABOLITE NAME", style=style)
		sheet.write(0, 2, "METABOLITE COMPARTMENT", style=style)
		sheet.write(0, 3, "DEM", style=style)
		sheet.write(0, 4, "DEM AFTER FVA", style=style)
		i = 1
		for metabolite in metabolites:
			sheet.write(i, 0, metabolite.id)
			sheet.write(i, 1, metabolite.name)
			sheet.write(i, 2, metabolite.compartment)
			if metabolite.id in aux_dem_init:
				sheet.write(i, 3, "TRUE")
			if metabolite.id in aux_dem_fva:
				sheet.write(i, 4, "TRUE")
			i = i + 1


	def __aux_spreadsheet_write_chokepoints_genes(self, state):
		chokepoints_dict = {}
		if state.chokepoints() is not None:
			for (reaction, metabolite) in state.chokepoints():
				reaction = reaction.id
				metabolite = metabolite.id
				if reaction in chokepoints_dict:
					chokepoints_dict[reaction].append(metabolite)
				else:
					chokepoints_dict[reaction] = [metabolite]
		genes_reactions_dict = {}
		if state.essential_genes_reactions() is not None:
			for reaction in state.essential_genes_reactions().keys():
				for gen in state.essential_genes_reactions()[reaction]:
					reaction = reaction.id
					gen = gen.id
					if reaction in genes_reactions_dict:
						genes_reactions_dict[reaction].append(gen)
					else:
						genes_reactions_dict[reaction] = [gen]
		essential_reactions = {}
		if state.essential_reactions() is not None:
			for reaction, growth in state.essential_reactions().items():
				essential_reactions[reaction.id] = growth

		return (chokepoints_dict, genes_reactions_dict, essential_reactions)


	def spreadsheet_write_chokepoints_genes(self, sheet_name, state_initial, state_dem, state_fva, state_fva_dem):
		"""

		Args:
			sheet_name ():

		Returns:

		"""
		if self.__spreadsheet is None:
			self.spreadsheet_init()

		style = xlwt.XFStyle()
		font = xlwt.Font()
		font.bold = True
		style.font = font

		sheet = self.__spreadsheet.add_sheet(sheet_name)
		sheet.write(1, 0, "REACTION ID", style=style)
		sheet.write(0, 2, "INITIAL MODEL", style=style)
		sheet.write(0, 6, "MODEL AFTER D.E.M. REMOVED", style=style)
		sheet.write(0, 10, "MODEL AFTER F.V.A.", style=style)
		sheet.write(0, 14, "MODEL AFTER F.V.A. AND D.E.M. REMOVED", style=style)

		sheet.write(1, 2, "CHOKEPOINT", style=style)
		sheet.write(1, 3, "ESSENTIAL GENE REACTION", style=style)
		sheet.write(1, 4, "OBJECTIVE VALUE", style=style)
		sheet.write(1, 6, "CHOKEPOINT", style=style)
		sheet.write(1, 7, "ESSENTIAL GENE REACTION", style=style)
		sheet.write(1, 8, "OBJECTIVE VALUE", style=style)
		sheet.write(1, 10, "CHOKEPOINT", style=style)
		sheet.write(1, 11, "ESSENTIAL GENE REACTION", style=style)
		sheet.write(1, 12, "OBJECTIVE VALUE", style=style)
		sheet.write(1, 14, "CHOKEPOINT", style=style)
		sheet.write(1, 15, "ESSENTIAL GENE REACTION", style=style)
		sheet.write(1, 16, "OBJECTIVE VALUE", style=style)

		models = [state_initial, state_dem, state_fva, state_fva_dem]
		j = 2
		reactions = state_initial.reactions()
		reactions.sort(key=self.__id)
		for model in models:
			i = 2
			(chokepoints, genes_reactions, essential_reactions) = self.__aux_spreadsheet_write_chokepoints_genes(model)
			for reaction_obj in reactions:
				reaction = reaction_obj.id
				if j == 2:
					sheet.write(i, 0, reaction)
					sheet.write(i, 1, " ")
				if reaction in chokepoints.keys():
					sheet.write(i, j, "TRUE")
				if reaction in genes_reactions.keys():
					sheet.write(i, j + 1, "TRUE")
				if reaction in essential_reactions.keys():
					sheet.write(i, j + 2, essential_reactions[reaction])
				i = i + 1
			j = j + 4


	def __aux_spreadsheet_get_data(self, state):
		dem = []
		if state.dem() is not None:
			metabolites = list(itertools.chain.from_iterable(state.dem().values()))
			dem = [m.id for m in metabolites]

		chokepoints_dict = {}
		if state.chokepoints() is not None:
			for (reaction, metabolite) in state.chokepoints():
				reaction = reaction.id
				metabolite = metabolite.id
				if reaction in chokepoints_dict:
					chokepoints_dict[reaction].append(metabolite)
				else:
					chokepoints_dict[reaction] = [metabolite]
		genes_reactions_dict = {}
		if state.essential_genes_reactions() is not None:
			for reaction in state.essential_genes_reactions().keys():
				for gen in state.essential_genes_reactions()[reaction]:
					reaction = reaction.id
					gen = gen.id
					if reaction in genes_reactions_dict:
						genes_reactions_dict[reaction].append(gen)
					else:
						genes_reactions_dict[reaction] = [gen]
		essential_reactions = {}
		if state.essential_reactions() is not None:
			for reaction, growth in state.essential_reactions().items():
				essential_reactions[reaction.id] = growth

		essential_genes = []
		if state.essential_genes() is not None:
			essential_genes = [g.id for g in state.essential_genes()]

		er_g0 = []
		er_g005 = []
		objective_value = state.objective_value()
		if state.objective_value() is not None:
			if not isnan(state.objective_value()):
				objective_value_005 = state.objective_value() * 0.05
			else:
				objective_value_005 = 0
			for reaction, value in essential_reactions.items():
				if not isnan(objective_value):
					if isnan(value) or value == 0:
						er_g0.append(reaction)
						er_g005.append(reaction)
					elif value < objective_value_005:
						er_g005.append(reaction)
				else:
					er_g0.append(reaction)
					er_g005.append(reaction)
		
		reversibles = state.reversible_reactions()

		dead_reactions = state.dead_reactions()

		return [dem, list(chokepoints_dict.keys()), essential_genes, genes_reactions_dict, essential_reactions, er_g0, er_g005, reversibles, dead_reactions]


	#
	def __aux_spreadsheet_write_two_model_summary(self, sheet, style, x, y, title, list_model_1, list_model_2):
		sheet.write(x, y, title, style=style)
		sheet.write(x + 2, y + 1, "Dead-end metabolites", style=style)
		sheet.write(x + 3, y + 1, "Chokepoint reactions", style=style)
		sheet.write(x + 4, y + 1, "Essential genes", style=style)
		sheet.write(x + 5, y + 1, "Essential genes reactions", style=style)
		sheet.write(x + 6, y + 1, "Essential reactions (objective=0)", style=style)
		sheet.write(x + 7, y + 1, "Essential reactions (objective < 5% max objective)", style=style)
		sheet.write(x + 8, y + 1, "Reversible reactions", style=style)
		sheet.write(x + 9, y + 1, "Dead reactions", style=style)
		sheet.write(x + 1, y + 2, "Before", style=style)
		sheet.write(x + 1, y + 3, "After", style=style)
		sheet.write(x + 1, y + 5, "Only before", style=style)
		sheet.write(x + 1, y + 6, "Intersection", style=style)
		sheet.write(x + 1, y + 7, "Only after", style=style)

		for i in range(0, len(list_model_1)):
			sheet.write(x + i + 2, y + 2, len(list_model_1[i]))
			sheet.write(x + i + 2, y + 3, len(list_model_2[i]))

		for i in range(0, len(list_model_1)):
			if isinstance(list_model_1[0], list):
				d1 = set(list_model_1[i])
				d2 = set(list_model_2[i])
				sheet.write(x + i + 2, y + 5, len(d1.difference(d2)))
				sheet.write(x + i + 2, y + 6, len(d1.intersection(d2)))
				sheet.write(x + i + 2, y + 7, len(d2.difference(d1)))
			else:
				d1 = list_model_1[i].keys()
				d2 = list_model_2[i].keys()
				sheet.write(x + i + 3, y + 5, len(d1 - d2))
				sheet.write(x + i + 3, y + 6, len(d1 & d2))
				sheet.write(x + i + 3, y + 7, len(d2 - d1))


	def sets_intersection(self, set_list):
		list_non_void_sets = []
		for st in set_list:
			if len(st) != 0:
				list_non_void_sets.append(st)
		if list_non_void_sets == []:
			return set()
		else:
			aux = list_non_void_sets[0]
			for i in range(1, len(list_non_void_sets)):
				aux = aux.intersection(list_non_void_sets[i])
			return aux


	def sets_union(self, set_list):
		if set_list == []:
			return set()
		else:
			aux = set_list[0]
			for i in range(1, len(set_list)):
				aux = aux.union(set_list[i])
			return aux


	def __aux_write_sets(self, sheet,  x, y, ls):
		for i in range(len(ls)):
			st = ls.pop(i)
			if len(st) != 0:
				val = st.difference(self.sets_union(ls))
				sheet.write(x, y + i + 2, len(val))
			ls.insert(i, st)
		val = self.sets_intersection(ls)
		sheet.write(x, y + 6, len(val))
		val = self.sets_union(ls)
		sheet.write(x, y + 7, len(val))

	def __aux_spreadsheet_write_four_model_summary(self, sheet, style, x, y, title, model_results):
		sheet.write(x, y, title, style=style)
		sheet.write(x + 2, y + 1, "Dead-end metabolites", style=style)
		sheet.write(x + 3, y + 1, "Chokepoint reactions", style=style)
		sheet.write(x + 4, y + 1, "Essential genes", style=style)
		sheet.write(x + 5, y + 1, "Essential genes reactions", style=style)
		sheet.write(x + 6, y + 1, "Essential reactions (objective=0)", style=style)
		sheet.write(x + 7, y + 1, "Essential reactions (objective < 5% max objective)", style=style)
		sheet.write(x + 8, y + 1, "Reversible reactions", style=style)
		sheet.write(x + 1, y + 2, "Only: initial model", style=style)
		sheet.write(x + 1, y + 3, "Only: model without D.E.M.", style=style)
		sheet.write(x + 1, y + 4, "Only: model updated with F.V.A.", style=style)
		sheet.write(x + 1, y + 5, "Only: model with F.V.A. without D.E.M.", style=style)
		sheet.write(x + 1, y + 6, "Intersection (only non void sets)", style=style)
		sheet.write(x + 1, y + 7, "Union 4 models", style=style)

		# turn data to sets to operate with
		for model in model_results:
			for i in range(0, len(model)):
				if isinstance(model[i], list):
					model[i] = set(model[i])
				else:
					# instance dict
					model[i] = set(model[i].keys())

		for i in range(0, len(model)):
			set_list = [model_results[0][i], model_results[1][i], model_results[2][i], model_results[3][i]]
			self.__aux_write_sets(sheet, x + i + 2, y, set_list)

	#
	def __aux_write_four_model_two_summary(self, sheet, style, x, y, title, subtitle1, list_set1, subtitle2, list_set2, subtitle3=None, list_set3=None):

		sheet.write(x, y, title, style=style)
		sheet.write(x + 2, y + 1, subtitle1, style=style)
		sheet.write(x + 3, y + 1, subtitle2, style=style)
		if subtitle3 is None:
			sheet.write(x + 4, y + 1, "Intersection", style=style)
			sheet.write(x + 5, y + 1, "Union", style=style)
		else:
			sheet.write(x + 4, y + 1, subtitle3, style=style)
			sheet.write(x + 5, y + 1, "Intersection", style=style)
			sheet.write(x + 6, y + 1, "Union", style=style)
		sheet.write(x + 1, y + 2, "Initial model", style=style)
		sheet.write(x + 1, y + 3, "Model without D.E.M.", style=style)
		sheet.write(x + 1, y + 4, "Model updated with F.V.A.", style=style)
		sheet.write(x + 1, y + 5, "Model with F.V.A. without D.E.M.", style=style)

		i = 0
		for aux_y in range(2, 6):
			if subtitle3 is None:
				if list_set1[i] is not None and list_set2[i] is not None:
					sheet.write(x + 2, aux_y, len(list_set1[i].difference(list_set2[i])))
					sheet.write(x + 3, aux_y, len(list_set2[i].difference(list_set1[i])))
					sheet.write(x + 4, aux_y, len(list_set1[i].intersection(list_set2[i])))
					sheet.write(x + 5, aux_y, len(list_set1[i].union(list_set2[i])))
				else:
					for aux_x in range(2, 6):
						sheet.write(x + aux_x, aux_y, "Nan")
			else:
				if list_set1[i] is not None and list_set2[i] is not None and list_set3[i] is not None:
					sheet.write(x + 2, aux_y, len(list_set1[i].difference(list_set2[i].union(list_set3[i]))))
					sheet.write(x + 3, aux_y, len(list_set2[i].difference(list_set1[i].union(list_set3[i]))))
					sheet.write(x + 4, aux_y, len(list_set3[i].difference(list_set2[i].union(list_set1[i]))))
					sheet.write(x + 5, aux_y, len( list_set1[i].intersection(list_set2[i]).intersection(list_set3[i]) ) )
					sheet.write(x + 6, aux_y, len( list_set1[i].union(list_set2[i]).union(list_set3[i]) ) )
				else:
					for aux_x in range(2, 7):
						sheet.write(x + aux_x, aux_y, "Nan")
			i = i + 1

	def spreadsheet_write_summary(self, sheet_name, state_initial, state_dem, state_fva, state_fva_dem):
		if self.__spreadsheet is None:
			self.spreadsheet_init()

		style = xlwt.XFStyle()
		font = xlwt.Font()
		font.bold = True
		style.font = font

		sheet = self.__spreadsheet.add_sheet(sheet_name)


		list_model_initial_results = self.__aux_spreadsheet_get_data(state_initial)
		list_model_fva_results = self.__aux_spreadsheet_get_data(state_fva)
		list_model_dem_results = self.__aux_spreadsheet_get_data(state_dem)
		list_model_fvadem_results = self.__aux_spreadsheet_get_data(state_fva_dem)
		models = [list_model_initial_results, list_model_dem_results, list_model_fva_results, list_model_fvadem_results]

		# remove essential_reactions list with objective values
		for m in models:
			del m[4]

		list_metabolites = [len(state_initial.metabolites()), len(state_dem.metabolites()), len(state_fva.metabolites()), len(state_fva_dem.metabolites())]
		list_reactions = [len(state_initial.reactions()), len(state_dem.reactions()), len(state_fva.reactions()), len(state_fva_dem.reactions())]

		# 1: list(chokepoints_dict.keys()),
		# 3: genes_reactions_dict, 
		# 5: er_g0
		reactions_data = []
		for i in [1,3,5]:
			rdl = [list_model_initial_results[i], list_model_dem_results[i], list_model_fva_results[i], list_model_fvadem_results[i]]
			for j in range(len(rdl)):
				if isinstance(rdl[j], list):
					rdl[j] = set(rdl[j])
				else:
					rdl[j] = set(rdl[j].keys())
			reactions_data.append(rdl)


		x = 0
		y = 0
		sheet.write(x, y, "MODELS DATA", style=style)
		sheet.write(x + 2, y + 1, "Metabolites", style=style)
		sheet.write(x + 3, y + 1, "Reactions", style=style)
		sheet.write(x + 1, y + 2, "Initial model", style=style)
		sheet.write(x + 1, y + 3, "Model without D.E.M.", style=style)
		sheet.write(x + 1, y + 4, "Model updated with F.V.A.", style=style)
		sheet.write(x + 1, y + 5, "Model with F.V.A. without D.E.M.", style=style)
		y = 2
		for m, r in zip(list_metabolites, list_reactions):
			sheet.write(x + 2, y, m)
			sheet.write(x + 3, y, r)
			y = y + 1

		self.__aux_spreadsheet_write_two_model_summary(sheet, style, 7, 0, "INITIAL MODEL BEFORE AND AFTER F.V.A.", list_model_initial_results, list_model_fva_results)
		self.__aux_spreadsheet_write_two_model_summary(sheet, style, 18, 0, "INITIAL MODEL BEFORE AND AFTER D.E.M. REMOVAL", list_model_initial_results, list_model_dem_results)
		self.__aux_spreadsheet_write_two_model_summary(sheet, style, 29, 0, "MODEL FIRST UPDATED WITH F.V.A. BEFORE AND AFTER D.E.M. REMOVAL", list_model_fva_results, list_model_fvadem_results)

		self.__aux_spreadsheet_write_four_model_summary(sheet, style, 40, 0, "FOUR MODEL COMPARISON", models)

		self.__aux_write_four_model_two_summary(sheet, style, 51, 0, "CHOKEPOINT REACTIONS - ESSENTIAL GENES REACTIONS", "Only chokepoint reaction", reactions_data[0], "Only essential genes reactions", reactions_data[1])
		self.__aux_write_four_model_two_summary(sheet, style, 58, 0, "CHOKEPOINT REACTIONS - ESSENTIAL REACTIONS (objectite value < 5% max objective value)", "Only chokepoints", reactions_data[0], "Only essential reactions", reactions_data[2])
		self.__aux_write_four_model_two_summary(sheet, style, 66, 0, "ESSENTIAL GENES REACTIONS - ESSENTIAL REACTIONS (objectite value < 5% max objective value)", "Only essential genes reactions",  reactions_data[1], "Only essential reactions", reactions_data[2])

		self.__aux_write_four_model_two_summary(sheet, style, 74, 0, "CHOKEPOINT REACTIONS - ESSENTIAL GENES REACTIONS - ESSENTIAL REACTIONS (objectite value < 5% max objective value)", "Only chokepoints reactions", reactions_data[0], "Only essential genes reactions",  reactions_data[1], "Only essential reactions", reactions_data[2])

	def spreadsheet_save_file(self, filename):
		""" Saves the xlwt Workbook '__spreadsheet' to a valid file.

		Args:
			filename (): output file.
		"""
		if self.__spreadsheet is None:
			self.spreadsheet_init()

		if (filename[-4:] == ".xls") or (filename[-4:] == ".ods") or (filename[-5:] == ".xlsx"):
			self.__spreadsheet.save(filename)
		else:
			raise Exception("File must be: '.xls', '.ods', '.xlsx'")



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="findCPcore",
    version="0.0.9",
    author="Alex Oarga",
    author_email="alex718123@gmail.com",
    description="findCP core package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/findCP/findCPcore",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "cobra==0.16.0",
        "depinfo",
        "future",
        "mpmath",
        "numpy>=1.13",
        "optlang>=1.4.2",
        "pandas>=0.17.0",
        "python-dateutil",
        "python-libsbml",
        "python-libsbml-experimental==5.18.0",
        "pytz",
        "ruamel.yaml>=0.15",
        "setuptools",
        "six",
        "swiglpk",
        "sympy",
        "tabulate",
        "xlwt==1.3.0",
	"six",
    ],
    python_requires='>=3.5',
)



from findCPcore.AbstractMetabolicModel import *
from findCPcore.MetabolicModel import *
from findCPcore.CobraMetabolicModel import *
from findCPcore.State import *
from findCPcore.Spreadsheet import *
from findCPcore.FacadeUtils import *
from findCPcore.Facade import *
from findCPcore.FacadeThread import *

